Package: scClassify
Type: Package
Title: scClassify: single-cell Hierarchical Classification
Version: 1.24.0
Author: Yingxin Lin
Maintainer: Yingxin Lin <yingxin.lin@sydney.edu.au>
Description: scClassify is a multiscale classification framework for single-cell RNA-seq data based on ensemble learning and cell type hierarchies, enabling sample size estimation required for accurate cell type classification and joint classification of cells using multiple references.
License: GPL-3
Encoding: UTF-8
LazyData: false
Depends: R (>= 4.0)
Imports: S4Vectors, limma, ggraph, igraph, methods, cluster,
        minpack.lm, mixtools, BiocParallel, proxy, proxyC, Matrix,
        ggplot2, hopach, diptest, mgcv, stats, graphics, statmod, Cepo
RoxygenNote: 7.1.1
Suggests: knitr, rmarkdown, BiocStyle, pkgdown
VignetteBuilder: knitr
biocViews: SingleCell, GeneExpression, Classification
BugReports: https://github.com/SydneyBioX/scClassify/issues
git_url: https://git.bioconductor.org/packages/scClassify
git_branch: RELEASE_3_23
git_last_commit: 8f9d87a
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.23
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:27:01 UTC; biocbuild

# Generated by roxygen2: do not edit by hand

export(cellTypeTrain)
export(cellTypeTree)
export(features)
export(getN)
export(learningCurve)
export(model)
export(modelweights)
export(name)
export(plotCellTypeTree)
export(predict_scClassify)
export(predict_scClassifyJoint)
export(runHOPACH)
export(runSampleCal)
export(scClassify)
export(scClassifyTrainModelList)
export(train_scClassify)
exportClasses(scClassifyTrainModel)
exportClasses(scClassifyTrainModelList)
import(ggplot2)
import(igraph)
import(limma)
import(statmod)
importClassesFrom(S4Vectors,DataFrame)
importClassesFrom(S4Vectors,SimpleList)
importClassesFrom(S4Vectors,character_OR_NULL)
importFrom(BiocParallel,SerialParam)
importFrom(BiocParallel,bplapply)
importFrom(Cepo,Cepo)
importFrom(Cepo,topGenes)
importFrom(S4Vectors,DataFrame)
importFrom(S4Vectors,SimpleList)
importFrom(cluster,pam)
importFrom(ggraph,geom_edge_diagonal)
importFrom(ggraph,geom_node_point)
importFrom(ggraph,geom_node_text)
importFrom(ggraph,ggraph)
importFrom(graphics,abline)
importFrom(graphics,plot)
importFrom(graphics,text)
importFrom(hopach,as.hdist)
importFrom(hopach,distancevector)
importFrom(hopach,hdist)
importFrom(hopach,improveordering)
importFrom(hopach,is.hdist)
importFrom(hopach,vectmatrix)
importFrom(limma,eBayes)
importFrom(limma,lmFit)
importFrom(methods,as)
importFrom(methods,is)
importFrom(methods,new)
importFrom(methods,slot)
importFrom(mgcv,gam)
importFrom(minpack.lm,nlsLM)
importFrom(mixtools,normalmixEM)
importFrom(proxy,as.dist)
importFrom(proxy,dist)
importFrom(proxyC,simil)
importFrom(stats,coef)
importFrom(stats,cutree)
importFrom(stats,dnorm)
importFrom(stats,lm)
importFrom(stats,median)
importFrom(stats,na.omit)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,quantile)
importFrom(stats,uniroot)

# scClassify: hierarchical classification of cells

<img src="man/figures/scClassifySticker.png" align="right" width="200"/>

Single cell classification via cell-type hierarchies based on ensemble learning and sample size estimation.


## Installation


Install Bioconductor packages `S4Vectors`, `hopach` and `limma` packages using `BiocManager`:

```r
# install.packages("BiocManager")
BiocManager::install(c("S4Vectors", "hopach", "limma"))
```

Then install the latest `scClassify` using `devtools` (For R >= 4.0):

```r
library(devtools)
devtools::install_github("SydneyBioX/scClassify")
```

For R >= 3.6, install `scClassify(v0.2.3)` via 

```
devtools::install_github("SydneyBioX/scClassify@085c72f")
```

## Vignette and Shiny app

You can find the vignette at this website (https://sydneybiox.github.io/scClassify/index.html):

+ scClassify Model Building and Prediction: https://sydneybiox.github.io/scClassify/articles/scClassify.html
+ Sample size calculation: https://sydneybiox.github.io/scClassify/articles/webOnly/sampleSizeCal.html
+ Performing scClassify using pretrained models: https://sydneybiox.github.io/scClassify/articles/pretrainedModel.html


Also, you can find our interactive shiny application (beta) at this website:
http://shiny.maths.usyd.edu.au/scClassify.



## Pretrained models

Currently available pre-trained scClassify models (in `scClassifyTrainModel` class)

|          Tissue        | Organism | Training Data | Accession |  Summary                                | Download `.rds`  | Gene Name Format  |
| :--: | :--: | :--: | :--: | :--: | :--: | :--: |
| Primary visual cortex  |   mouse  |  Tasic (2018) | [GSE115746](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE115746)    | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Tasic2018_mouseNeuronal.html)|    [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainTasicV2resClass.rds)    | Mm Gene Symbol|
| Primary visual cortex  |   mouse  |  Tasic (2016) | [GSE71585](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE71585)    | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Tasic2016_mouseNeuronal.html)|  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainTasicV1resClass.rds)   | Mm Gene Symbol|
| Visual cortex          |   mouse  |  Hrvatin      | [GSE102827](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE102827)  |   [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Hrvatin_mouseNeuronal.html)        |  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainHrvatinresClass.rds)  | Mm Gene Symbol|
|    Lung                |   mouse  |  Cohen        |   [GSE119228](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE119228)        |    [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Cohen_mouseLung.html)        |  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainCohenresClass.rds)    | Mm Gene Symbol|
|    Kidney              |   mouse  |  Park         |  [GSE107585](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE107585)         | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Park_mouseKidney.html)          |   [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainParkresClass.rds)     | Mm Gene Symbol|
|    Liver               |   human  |  MacParland   | [GSE115469](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE115469)          | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/MacParlandres_humanLiver.html)   | [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainMacParlandresClass.rds)  | Hs Gene Symbol|
|    Liver               |   human  |  Aizarani     |  [GSE124395](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE124395)         | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Aizarani_humanLiver.html)        |  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainAizaraniresClass.rds) | Hs Gene Symbol|
|    Pancreas            |   human  |  Xin          |   [GSE81608](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE81608)         |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Xin_humanPancreas.html)          |  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainXinClass.rds)                   | Hs Gene Symbol|
|    Pancreas            |   human  |  Wang         |   [GSE83139](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE83139)        |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Wang_humanPancreas.html)          |     [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainWangClass.rds)                | Hs Gene Symbol|
|    Pancreas            |   human  |  Lawlor       |   [GSE86469](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE86469)         |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Lawlor_humanPancreas.html)           |     [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainLawlorClass.rds)                | Hs Gene Symbol|
|    Pancreas            |   human  |  Segerstolpe  |   [E-MTAB-5061](https://www.ebi.ac.uk/arrayexpress/E-MTAB-5061)        |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Segerstolpe_humanPancreas.html)           |  [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainSegerstolpeClass.rds)                  | Hs Gene Symbol|
|    Pancreas            |   human  |  Muraro       |    [GSE85241](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE85241)         | [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Muraro_humanPancreas.html)           |     [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainMuraroClass.rds)               | Hs Gene Symbol|
|    Pancreas            |   human  |  Baron        |   [GSE84133](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE84133)         |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Baron_humanPancreas.html)          |   [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainBaronClass.rds)               | Hs Gene Symbol|
|    Pancreas            |   human  |  joint        |    -     |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Joint_humanPancreas.html)          |        [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/jointPancreasClass.rds)          | Hs Gene Symbol|
|    Melanoma            |   human  |  Li           | [GSE123139](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE123139)          |   [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Li_humanMelanoma.html)         |    [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainLiresClass.rds)     | Hs Gene Symbol|
|    PBMC                |   human  |  Ding (joint) |    -   |  [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/Joint_humanPBMC.html)   |           [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/jointPBMCClass.rds)        | Mm EMSEMBL ID|
|    Tabula Muris        |   mouse  |  Tabula Muris | [GSE109774](https://www.ncbi.nlm.nih.gov/geo/query/acc.cgi?acc=GSE109774)           |   [link](https://SydneyBioX.github.io/scClassify/articles/webOnly/TabulaMuris.html)        |    [link](http://www.maths.usyd.edu.au/u/yingxinl/wwwnb/scClassify/trainTMresClass.rds)              | Mm Gene Symbol|



## Contact us

If you have any enquiries, especially about performing `scClassify` to classify your cells or to build your own models, please contact <yingxin.lin@sydney.edu.au> or <bioinformatics@maths.usyd.edu.au>.


## Reference

**scClassify: sample size estimation and multiscale classification of cells using single and multiple reference**

Yingxin Lin, Yue Cao, Hani Kim, Agus Salim, Terence Speed, David Lin, Pengyi Yang† & Jean Yang†. ***Molecular Systems Biology***, 2020, 16, e9389. [Full Text](https://doi.org/10.15252/msb.20199389); [BioC R package](http://www.bioconductor.org/packages/release/bioc/html/scClassify.html)

## ----include = FALSE----------------------------------------------------------
knitr::opts_chunk$set(
    collapse = TRUE,
    warning = FALSE,
    message = FALSE,
    comment = "#>"
)

## ----eval = FALSE-------------------------------------------------------------
# # installation of scClassify
# if (!requireNamespace("BiocManager", quietly = TRUE)) {
#     install.packages("BiocManager")
# }
# BiocManager::install("scClassify")

## ----setup--------------------------------------------------------------------
library("scClassify")
data("scClassify_example")
xin_cellTypes <- scClassify_example$xin_cellTypes
exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
wang_cellTypes <- scClassify_example$wang_cellTypes
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
exprsMat_xin_subset <- as(exprsMat_xin_subset, "dgCMatrix")
exprsMat_wang_subset <- as(exprsMat_wang_subset, "dgCMatrix")

## -----------------------------------------------------------------------------
table(xin_cellTypes)
table(wang_cellTypes)

## -----------------------------------------------------------------------------
scClassify_res <- scClassify(exprsMat_train = exprsMat_xin_subset,
                             cellTypes_train = xin_cellTypes,
                             exprsMat_test = list(wang = exprsMat_wang_subset),
                             cellTypes_test = list(wang = wang_cellTypes),
                             tree = "HOPACH",
                             algorithm = "WKNN",
                             selectFeatures = c("limma"),
                             similarity = c("pearson"),
                             returnList = FALSE,
                             verbose = FALSE)


## ----warning=FALSE------------------------------------------------------------
scClassify_res$trainRes
plotCellTypeTree(cellTypeTree(scClassify_res$trainRes))

## -----------------------------------------------------------------------------
table(scClassify_res$testRes$wang$pearson_WKNN_limma$predRes, wang_cellTypes)

## -----------------------------------------------------------------------------
scClassify_res_ensemble <- scClassify(exprsMat_train = exprsMat_xin_subset,
                                      cellTypes_train = xin_cellTypes,
                                      exprsMat_test = list(wang = exprsMat_wang_subset),
                                      cellTypes_test = list(wang = wang_cellTypes),
                                      tree = "HOPACH",
                                      algorithm = "WKNN",
                                      selectFeatures = c("limma"),
                                      similarity = c("pearson", "cosine"),
                                      weighted_ensemble = FALSE,
                                      returnList = FALSE,
                                      verbose = FALSE)


## -----------------------------------------------------------------------------
table(scClassify_res_ensemble$testRes$wang$pearson_WKNN_limma$predRes,
      scClassify_res_ensemble$testRes$wang$cosine_WKNN_limma$predRes)

## -----------------------------------------------------------------------------
table(scClassify_res_ensemble$testRes$wang$ensembleRes$cellTypes, 
      wang_cellTypes)

## -----------------------------------------------------------------------------
trainClass <- train_scClassify(exprsMat_train = exprsMat_xin_subset,
                               cellTypes_train = xin_cellTypes,
                               selectFeatures = c("limma", "BI"),
                               returnList = FALSE
)

## -----------------------------------------------------------------------------
trainClass

## -----------------------------------------------------------------------------
sessionInfo()

# functions get ensemble results



# function to transfer the error
alpha <- function(e) {
    e[e < 0.001] <- 0.001
    e[e > 0.999] <- 0.999
    log((1 - e)/e)
}



getTrainWeights <- function(trainRes) {
    errClass_train <- do.call(rbind, lapply(trainRes, function(x)
        table(x$classifyRes)/length(x$classifyRes)))
    weighted_train <- alpha(1 - (errClass_train[, 1] + errClass_train[, 2]))
    return(weighted_train)
}


# function to get the ensemble res
getEnsembleRes <- function(testRes, weighted_train,
                           exclude = NULL, weighted_ensemble = TRUE) {
    #
    #    trainRes <- testRes$train
    # Get the weighted
    if (weighted_ensemble) {
        if (!is.null(exclude)) {
            keep_method <- !grepl(paste(exclude, collapse = "|"),
                                  names(weighted_train))
            weighted_train <- weighted_train[keep_method]
        }
        
    } else {
        weighted_train <- NULL
    }
    
    ensembleRes <- calEnsembleRes(testRes,
                                  exclude = exclude,
                                  weight = weighted_train)
    return(ensembleRes)
    
}

# Function to combine the ensemble res

calEnsembleRes <- function(res, exclude = NULL, weight = NULL){
    
    
    if (!is.null(exclude)) {
        keep_method <- !grepl(paste(exclude,collapse = "|"), names(res))
        res <- res[keep_method]
        weight <- weight[keep_method]
    }
    

    ensembleResMat <- do.call(cbind, lapply(res, function(x) x$predRes))
    
    if (is.null(weight)) {
        weight <- rep(1, ncol(ensembleResMat))
    }
    
    ensembleRes <- apply(ensembleResMat, 1, function(x) {
        names(x) <- colnames(ensembleResMat)
        keep <- rep(TRUE, length(x))
        if (sum(keep) == 0) {
            data.frame(cellTypes = "unassigned", scores = 0)
        }else{
            getResByWeights(x[keep], weight[keep])
        }
        
    })
    
    ensembleRes <- do.call(rbind, ensembleRes)
    return(ensembleRes)
}




# function to calculate the weight for each base classifier

getResByWeights <- function(res, weight) {
    resType <- unique(res)
    if (length(resType) == 1) {
        final <- resType
        scores <- 1
    } else {
        
        
        mat <- vapply(seq_len(length(resType)), function(i){
            ifelse(res %in% resType[i], 1, 0) * weight
        }, numeric(length(res)))
        
        colnames(mat) <- resType
        rownames(mat) <- names(res)
        mat_colMeans <- Matrix::colMeans(mat)
        final <- names(mat_colMeans)[which.max(mat_colMeans)]
        scores <- max(mat_colMeans)/sum(mat_colMeans)
    }
    
    return(data.frame(cellTypes = final, scores = scores))
}




# Function to summarise joint classification results

getJointRes <- function(predictResList, trainRes) {
    
    predictJointResList <- lapply(seq_along(predictResList), function(l) {
        
        if ("scClassifyTrainModel" %in% is(trainRes[[l]])) {
            weights_train <- modelweights(trainRes[[l]])
        } else {
            weights_train <- trainRes[[l]]$modelweights
        }
        
        getEnsembleRes(predictResList[[l]],
                       weights_train,
                       exclude = NULL,
                       weighted_ensemble = TRUE)
    })
    
    names(predictJointResList) <- names(predictResList)
    
    
    jointCellType <- do.call(cbind, lapply(predictJointResList, function(x)
        x$cellTypes))
    colnames(jointCellType) <- names(predictJointResList)
    
    
    jointCellType <- apply(jointCellType, 2, function(x) {
        len <- unlist(lapply(strsplit(x, "_"), length))
        x[len > 2] <- "intermediate"
        x
    })
    
    jointWeights <- do.call(cbind, lapply(predictJointResList, function(x)
        x$scores))
    colnames(jointWeights) <- names(predictJointResList)
    
    
    
    jointRes <- lapply(seq_len(nrow(jointCellType)), function(idx) {
        cell_res <- jointCellType[idx, ]
        names(cell_res) <- colnames(jointCellType)
        if (length(cell_res) == 2 & sum(cell_res == "unassigned") == 1) {
            data.frame(cellTypes = cell_res[cell_res != "unassigned"],
                       scores = 0.5)
        } else {
            getResByWeights(cell_res, jointWeights[idx, ])
        }
    })
    
    
    jointRes <- do.call(rbind, jointRes)
    
    jointRes <- data.frame(jointRes)
    
    return(jointRes)
}

#' @title Run sample size calculation for pilot data for reference dataset
#'
#' @param exprsMat A matrix of expression matrix of pilot dataset
#' (log-transformed, or normalised)
#' @param cellTypes A vector of cell types of pilot dataset
#' @param n_list A vector of integer indicates the sample size to run.
#' @param num_repeat An integer indicates the number of run for
#' each sample size will be repeated.
#' @param BPPARAM  A \code{BiocParallelParam} class object
#' from the \code{BiocParallel} package is used. Default is SerialParam().
#' @param subset_test A ogical input indicates whether we used a subset of data
#' (fixed number for each sample size)
#' to test instead of all remaining data. By default, it is FALSE.
#' @param num_test An integer indicates the size of the test data.
#' @param cellType_tree A list indicates the cell type tree (optional),
#' if it is NULL, the accuracy rate is calculate
#' based on the provided cellTypes.
#' @param level An integer indicates the accuracy rate is calculate
#' based on the n-th level from top of cell type tree.
#' If it is NULL (by default), it will be the bottom of the cell type tree.
#' It can not be larger than the total number of levels of the tree.
#' @param ... other parameter from scClassify
#'
#' @return A matrix of accuracy matrix, where columns corresponding to different
#' sample sizes, rows corresponding to the number of repetation.
#'
#' @examples
#' data("scClassify_example")
#' xin_cellTypes <- scClassify_example$xin_cellTypes
#' exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
#'
#' exprsMat_xin_subset <- as(exprsMat_xin_subset, "dgCMatrix")
#' set.seed(2019)
#' accMat <- runSampleCal(exprsMat_xin_subset,
#' xin_cellTypes,
#' n_list = seq(20, 100, 20),
#' num_repeat = 5, BPPARAM = BiocParallel::SerialParam())
#'
#' @importFrom BiocParallel bplapply SerialParam
#' @export


runSampleCal <- function(exprsMat,
                         cellTypes,
                         n_list = c(20, 40, 60, 80, 100, seq(200, 500, 100)),
                         num_repeat = 20,
                         level = NULL,
                         cellType_tree = NULL,
                         BPPARAM = BiocParallel::SerialParam(),
                         subset_test = FALSE,
                         num_test = NULL,
                         ...) {




    if (length(n_list) < 5) {
        stop("length of n_list provided is too short...
             wont be enough point to fit the learning curve")
    }




    # If there is an input of cell type tree,
    #relabelled the cell types based on input level

    if (!is.null(cellType_tree)) {
        if (is.null(level)) {
            level <- length(cellType_tree)
        } else {
            if (level <= 0 | level > length(cellType_tree)) {
                stop("The input of level is invalid")
            }

            level <- level + 1
        }

        cellTypes_ind <- cellType_tree[[level]][as.character(cellTypes)]
        cellTypes_relabelled <- vapply(seq_len(length(cellTypes_ind)),
           function(i) {
               paste(names(cellType_tree[[level]])[cellType_tree[[level]] %in%
                                                       cellTypes_ind[i]],
                     collapse = "_")
           }, character(1L))

        cellTypes <- cellTypes_relabelled
        cellType_tree_relabelled <- reLevelCellTypeTree(cellType_tree,
                                                        level = level)

    } else {
        cellType_tree_relabelled <- NULL
    }


    exprsMat <- as(exprsMat, "dgCMatrix")
    # n_list <- c(20, 40, 60, 80, 100, seq(200, 500, 100))

    res_sub <- list()

    for (i in seq_len(length(n_list))) {
        print(paste("n=",n_list[i]))
        res_sub[[i]] <- BiocParallel::bplapply(seq_len(num_repeat),
           function(x) {
               tryCatch({
                   l <- runSubSampling(exprsMat, cellTypes, n = n_list[[i]],
                                       subset_test = subset_test,
                                       num_test = num_test,
                                       cellType_tree = cellType_tree_relabelled,
                                       cutoff_method = "dynamic",
                                       prob_threshold = 0.6
                   )

                   table(l)/length(l)},
                   error = function(e){NULL})
           }, BPPARAM = BPPARAM)
        print(do.call(cbind, res_sub[[i]]))
        gc(reset = TRUE)
    }

    names(res_sub) <- n_list

    res_sub <- lapply(res_sub, function(x) unlist(lapply(x, "[[", "correct")))
    res_sub <- res_sub[!is.null(res_sub)]
    accuracy_mat <- do.call(cbind, res_sub)
    return(accuracy_mat)
}




# function to run subsampling


runSubSampling <- function(exprsMat,
                           cellTypes,
                           n = 50,
                           subset_test = FALSE,
                           num_test = 2000,
                           ...){


    # make sure the smallest training type has at least 3 cells.
    n_subset <- round(table(cellTypes)/length(cellTypes)*n)
    n_subset <- ifelse(n_subset < 3, 3, n_subset)

    trainIdx <- unlist(lapply(seq_len(length(n_subset)), function(x)
        sample(which(cellTypes == names(n_subset)[x]), n_subset[x])))

    exprsMat_train <- exprsMat[,trainIdx]
    cellTypes_train <- cellTypes[trainIdx]

    print(table(cellTypes_train))

    if (subset_test) {
        testIdx <- sample(seq_len(ncol(exprsMat))[-trainIdx], num_test)
        exprsMat_test <- exprsMat[, testIdx]
        cellTypes_test <- cellTypes[testIdx]
    }else{
        exprsMat_test <- exprsMat[, -trainIdx]
        cellTypes_test <- cellTypes[-trainIdx]
    }


    trainRes <- scClassify(exprsMat_train = exprsMat_train,
                           cellTypes_train = cellTypes_train,
                           exprsMat_test = list(
                               test = exprsMat_test),
                           cellTypes_test = list(
                               test = cellTypes_test),
                           ...)

    acc_cls <- trainRes$testRes$test$pearson_WKNN_limma$classifyRes
    return(acc_cls)
}



reLevelCellTypeTree <- function(cellType_tree, level) {

    cellTypes_newTree <- list()

    newCellTypeNames <- vapply(seq_len(max(cellType_tree[[level]])),
                               function(i) {
        paste(names(cellType_tree[[level]])[cellType_tree[[level]] == i],
              collapse = "_")
    }, character(1))

    cellTypes_newTree[[level]] <- seq_len(max(cellType_tree[[level]]))
    names(cellTypes_newTree[[level]]) <- newCellTypeNames

    for (i in seq_len((level - 1))) {

        cellTypes_newTree[[i]] <- vapply(seq_len(length(newCellTypeNames)),
         function(j)
             unique(cellType_tree[[i]][names(cellType_tree[[i]]) %in%
                                           unlist(strsplit(newCellTypeNames[j],
                                                           "_"))]),
         numeric(1)
        )
        names(cellTypes_newTree[[i]]) <- newCellTypeNames

    }

    return(cellTypes_newTree)
}

#' @importFrom methods new
#' @importFrom Cepo Cepo topGenes
#' @import limma


featureSelection <- function(exprsMat,
                             trainClass,
                             feature = c("limma", "DV", "DD", "chisq", "BI",
                                         "Cepo"),
                             topN = 50,
                             pSig = 0.001
){

    feature <- match.arg(feature, c("limma", "DV", "DD", "chisq", "BI", "Cepo"),
                         several.ok = FALSE)


    if (feature == "DV") {
        tt <- doDV(exprsMat, trainClass)
        tt <- lapply(tt, function(x)sort(x))
        res <- Reduce(union, lapply(tt, function(t)
            names(t)[seq_len(min(topN, sum(t < pSig)))]))
    } else if (feature == "DD") {
        tt <- doDD(exprsMat, trainClass)
        tt <- lapply(tt, function(x)sort(x))
        res <- Reduce(union, lapply(tt, function(t)
            names(t)[seq_len(min(topN, sum(t < pSig)))]))
    } else if (feature == "chisq") {
        tt <- doChisSquared(exprsMat, trainClass)
        tt <- lapply(tt, function(x)sort(x))
        res <- Reduce(union, lapply(tt, function(t)
            names(t)[seq_len(min(topN, sum(t < pSig)))]))
        #
    } else if (feature == "BI") {
        tt <- doBI(exprsMat, trainClass)
        tt <- lapply(tt, function(x)x)
        res <- Reduce(union, lapply(tt, function(t)
            names(t)[seq_len(topN)]))
    } else if (feature == "Cepo") {
        tt <- Cepo::Cepo(as.matrix(exprsMat), trainClass, exprsPct = 0.05)
        res <- Reduce(union, Cepo::topGenes(tt, n = topN))
    } else{
        tt <- doLimma(exprsMat, trainClass)
        res <- Reduce(union, lapply(tt, function(t)
            rownames(t[t$logFC > 0 & (t$meanPct.2 - t$meanPct.1) > 0.05 &
                           t$adj.P.Val < pSig,])[seq_len(topN)]))

    }

    return(res)
}


#' @importFrom limma eBayes lmFit
#' @importFrom methods new

doLimma <- function(exprsMat, cellTypes, exprs_pct = 0.05){

    cellTypes <- droplevels(as.factor(cellTypes))
    tt <- list()
    for (i in seq_len(nlevels(cellTypes))) {
        tmp_celltype <- (ifelse(cellTypes == levels(cellTypes)[i], 1, 0))
        design <- stats::model.matrix(~tmp_celltype)


        meanExprs <- do.call(cbind, lapply(c(0,1), function(i){
            Matrix::rowMeans(exprsMat[, tmp_celltype == i, drop = FALSE])
        }))

        meanPct <- do.call(cbind, lapply(c(0,1), function(i){
            Matrix::rowSums(exprsMat[, tmp_celltype == i,
                                     drop = FALSE] > 0)/sum(tmp_celltype == i)
        }))

        keep <- meanPct[,2] > exprs_pct

        y <- methods::new("EList")
        y$E <- exprsMat[keep, ]
        fit <- limma::lmFit(y, design = design)
        fit <- limma::eBayes(fit, trend = TRUE, robust = TRUE)
        tt[[i]] <- limma::topTable(fit, n = Inf, adjust.method = "BH", coef = 2)



        if (!is.null(tt[[i]]$ID)) {
            tt[[i]] <- tt[[i]][!duplicated(tt[[i]]$ID),]
            rownames(tt[[i]]) <- tt[[i]]$ID
        }

        tt[[i]]$meanExprs.1 <- meanExprs[rownames(tt[[i]]), 1]
        tt[[i]]$meanExprs.2 <- meanExprs[rownames(tt[[i]]), 2]
        tt[[i]]$meanPct.1 <- meanPct[rownames(tt[[i]]), 1]
        tt[[i]]$meanPct.2 <- meanPct[rownames(tt[[i]]), 2]
    }



    return(tt)


}




doDV <- function(exprsMat, cellTypes){


    cellTypes <- droplevels(as.factor(cellTypes))
    tt <- list()
    for (i in seq_len(nlevels(cellTypes))) {
        tmp_celltype <- (ifelse(cellTypes == levels(cellTypes)[i], 1, 0))

        meanPct <- do.call(cbind, lapply(c(0,1), function(i){
            Matrix::rowSums(exprsMat[,
                                     tmp_celltype == i,
                                     drop = FALSE] > 0)/sum(tmp_celltype == i)
        }))


        posNeg <- (meanPct[,2] - meanPct[,1]) > 0.05
        # print(sum(posNeg))
        exprsMat_filt <- exprsMat[posNeg,]
        tt[[i]] <- apply(exprsMat_filt, 1, function(x) {
            df <- data.frame(gene = x, cellTypes = as.factor(tmp_celltype))
            stats::bartlett.test(gene~cellTypes, df)$p.value
        })

        tt[[i]] <- stats::p.adjust(tt[[i]], method = "BH")
    }



    return(tt)


}

doDD <- function(exprsMat, cellTypes){

    cellTypes <- droplevels(as.factor(cellTypes))
    tt <- list()
    for (i in seq_len(nlevels(cellTypes))) {
        tmp_celltype <- ifelse(cellTypes == levels(cellTypes)[i], 1, 0)


        meanPct <- do.call(cbind, lapply(c(0,1), function(i){
            Matrix::rowSums(exprsMat[,
                                     tmp_celltype == i,
                                     drop = FALSE] > 0)/sum(tmp_celltype == i)
        }))

        posNeg <- (meanPct[,2] - meanPct[,1]) > 0.05
        # print(sum(posNeg))
        exprsMat_filt <- exprsMat[posNeg,]
        tt[[i]] <- apply(exprsMat_filt, 1, function(x) {
            x1 <- x[tmp_celltype == 0]
            x2 <- x[tmp_celltype == 1]
            stats::ks.test(x1, x2, alternative = "greater")$p.value
        })



        tt[[i]] <- stats::p.adjust(tt[[i]], method = "BH")
    }



    return(tt)


}



doChisSquared <- function(exprsMat, cellTypes, threshold = 1){


    cellTypes <- droplevels(as.factor(cellTypes))
    tt <- list()
    for (i in seq_len(nlevels(cellTypes))) {
        tmp_celltype <- (ifelse(cellTypes == levels(cellTypes)[i], 1, 0))


        zerosMat <- ifelse(exprsMat > threshold, 1, 0)

        tt[[i]] <- apply(zerosMat,1,  function(x){
            tab <- c()
            for (i in c(0,1)) {
                tmp <- factor(x[tmp_celltype == i], levels = c(0, 1))
                tab <- rbind(tab, table(tmp))
            }


            suppressWarnings(stats::chisq.test(tab)$p.value)


        })




        tt[[i]] <- stats::p.adjust(tt[[i]], method = "BH")
    }



    return(tt)


}






doBI <- function(exprsMat, cellTypes){
    # Select genes by bimodal index

    cellTypes <- droplevels(as.factor(cellTypes))
    tt <- list()
    for (i in seq_len(nlevels(cellTypes))) {
        tmp_celltype <- (ifelse(cellTypes == levels(cellTypes)[i], 1, 0))

        pi <- table(tmp_celltype)/length(tmp_celltype)

        agg_mean <- do.call(cbind, lapply(c(0,1), function(i){
            Matrix::rowMeans(exprsMat[, tmp_celltype == i, drop = FALSE])
        }))

        agg_sd2 <- do.call(cbind, lapply(c(0,1), function(i){
            apply(exprsMat[, tmp_celltype == i, drop = FALSE], 1, stats::var)
        }))

        bi <- abs(agg_mean[,2] - agg_mean[,1])/sqrt(pi[1]*agg_sd2[,1] +
                                                        pi[2]*agg_sd2[,2])

        bi <- unlist(bi)
        names(bi) <- rownames(exprsMat)
        bi <- bi[order(bi, decreasing = TRUE)]
        tt[[i]] <- bi
    }

    return(tt)


}

#' Create HOPACH tree
#'
#' A function generating HOPACH tree using the average expression matrix for
#' each cell type.
#'
#' @param data A matrix of average expression matrix
#'  (each row indicates the gene, each column indicates the cell type)
#' @param plot Indicate whether plot or not
#' @param kmax Integer between 1 and 9 specifying the maximum number of children
#' at each node in the tree.
#' @return Return a \code{list} where
#' \itemize{
#' \item{cutree_list}: A list indicates the hierarchical cell type tree
#' \item{plot}: A \code{ggplot} visualise the cell type tree
#' }
#' @author Yingxin Lin
#' @importFrom stats cutree median lm
#' @importFrom hopach vectmatrix hdist is.hdist improveordering distancevector
#' as.hdist
#' @importFrom methods as is
#' @importFrom cluster pam
#' @importFrom graphics plot text
#' @import statmod
#'
#' @examples
#'
#' data("scClassify_example")
#' wang_cellTypes <- factor(scClassify_example$wang_cellTypes)
#' exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
#' avgMat_wang <- apply(exprsMat_wang_subset, 1, function(x)
#' aggregate(x, list(wang_cellTypes), mean)$x)
#' rownames(avgMat_wang) <- levels(wang_cellTypes)
#' res_hopach <- runHOPACH(avgMat_wang)
#' res_hopach$plot
#'
#' @references van der Laan, M. J. and Pollard, K. S. (2003)
#' ‘A new algorithm for hybrid hierarchical clustering with
#' visualization and the bootstrap’,
#' Journal of Statistical Planning and Inference.
#' doi: 10.1016/S0378-3758(02)00388-9.
#' @export

runHOPACH <- function(data, plot= TRUE, kmax = 5) {

    # Compute the distance matrix using correlation
    dist <- distancematrix(data, d = "cor")

    # Run HOPACH function
    clustresult <- hopach(data, dmat = dist, kmax = kmax)


    # Rearrange HOPACH results
    final_labels <- strsplit(as.character(format(clustresult$final$labels,
                                                 scientific = FALSE)), "")

    # Get each level seperate
    level_list <- list()
    for (i in seq_len(nchar(clustresult$final$labels[1]))) {
        if (i != 1) {
            level_list[[i]] <- paste(level_list[[i - 1]],
                                     unlist(lapply(final_labels, "[[", i)),
                                     sep = "")
        }else{
            level_list[[i]] <- unlist(lapply(final_labels, "[[", i))
        }
    }



    # Generate cutree results
    cutree_list <- list()

    # First level (All ones)
    cutree_list[[1]] <- rep(1, length(rownames(data)))
    names(cutree_list[[1]]) <- rownames(data)

    cutree_list[2:(length(level_list) + 1)] <- lapply(level_list, function(x) {
        x <- as.numeric(as.factor(x))
        names(x) <- rownames(data)
        x
    })

    # Last levels will be all distinct numbers
    cutree_list[[length(cutree_list) + 1]] <-
        seq_len(length(clustresult$final$labels))
    names(cutree_list[[length(cutree_list)]]) <- rownames(data)

    # To plot using ggraph
    if (plot) {

        treePlot <- plotCellTypeTree(cutree_list)
        return(list(cutree_list = cutree_list, plot = treePlot))

    }else{
        return(list(cutree_list = cutree_list))
    }

}

#' To plot cell type tree
#'
#'
#' @param cutree_list A list indicates the hierarchical cell type tree
#' @param group_level Indicate whether plot or not
#'
#' @return A ggplot object visualising the HOPACH tree
#'
#' @importFrom ggraph geom_edge_diagonal geom_node_text geom_node_point ggraph
#' @import igraph
#'
#' @examples
#'
#' data("trainClassExample_xin")
#'
#' plotCellTypeTree(cellTypeTree(trainClassExample_xin))
#'
#' @export


plotCellTypeTree <- function(cutree_list, group_level = NULL) {

    cutree_list <- cutree_list

    if (is.null(group_level)) {
        group_level <- 3
    }

    E_list <- list()
    for (i in seq_len(length(cutree_list))) {
        if (i != 1) {
            parent <- cutree_list[[i - 1]]
            E_list[[i]] <- paste(paste(i - 1, parent, sep = ""),
                                 paste(i, cutree_list[[i]], sep = ""),
                                 sep = "_")
        }
    }

    V_list <- list()
    for (i in seq_len(length(cutree_list))) {
        V_list[[i]] <- paste(i, cutree_list[[i]], sep = "")
        names(V_list[[i]]) <- names(cutree_list[[i]])
    }


    g <- igraph::make_graph(unlist(strsplit(unique(unlist(lapply(E_list,
                                                                 sort))),
                                            "_")))



    ### Which level is used to color the node
    group_level <- group_level
    group_col <- (V_list[[group_level]])
    names(group_col) <- V_list[[length(cutree_list)]]


    V_group <- rep(NA, length(igraph::V(g)))
    names(V_group) <- names(V(g))
    V_group[V_list[[length(V_list)]]] <- group_col[V_list[[length(V_list)]]]
    V(g)$group <- V_group  # note only the bottom level are coloured


    ### Next, get the label for the bottom level

    V_name <- rep(NA, length(igraph::V(g)))
    names(V_name) <- names(V(g))
    V_name[V_list[[length(V_list)]]] <- names(V_list[[length(V_list)]])
    V(g)$name <- V_name




    treePlot <- ggraph::ggraph(g, layout = 'dendrogram') +
        ggraph::geom_edge_diagonal(colour = "black") +
        ggraph::geom_node_text(aes(vjust = 1, hjust = -0.2, label = name),
                               size = 2.7, alpha = 1) +
        ggraph::geom_node_point(aes(filter = leaf, alpha = 0.2, size = 3,
                                    colour = as.factor(V(g)$group))) +
        ggplot2::theme_void() +
        ggplot2::scale_colour_hue() +
        ggplot2::theme(legend.position = "none") +
        ggplot2::coord_flip() +
        NULL
    treePlot
}



#Hierarchical Ordered Partitioning and Collapsing Hybrid (HOPACH)#
#The codes here are originated from HOPACH pacakge with small edits
#
#Reference: ‘A new algorithm for hybrid hierarchical clustering with
# visualization and the bootstrap’,
# Journal of Statistical Planning and Inference.
# doi: 10.1016/S0378-3758(02)00388-9.

labelstomss <- function(labels, dist, khigh = 9, within = "med",
                        between = "med", hierarchical = TRUE) {
    if (!is.vector(labels))
        stop("First arg to labelstomss() must be a vector")

    p = methods::slot(dist, "Size")
    if (length(labels) != p)
        stop("Distance matrix and labels dimensions
             do not agree in labelstomss()")

    unlabels <- sort(unique(labels))
    k <- length(unlabels)
    ss <- NULL
    for (i in seq_len(k)) {
        labs <- (seq_len(p))[labels == unlabels[i]]
        pp <- length(labs)
        if (pp < 3)
            ss[i] <- NA
        else{
            dissvec <- methods::slot(dist[labs,labs], "Data")
            bestk <- silcheck(dissvec,min(khigh,(length(labs) - 1),
                                          na.rm = TRUE), diss = TRUE)[1]
            if (within == "med")
                ss[i] <- median(pam(dissvec, bestk,
                                    diss = TRUE)$silinfo$widths[, 3])
            if (within == "mean")
                ss[i] <- pam(dissvec,bestk,diss = TRUE)$silinfo$avg.width
        }
    }
    if (sum(is.na(ss)) == k)
        out <- NA
    else{
        if (hierarchical == TRUE) {
            parentlab <- trunc(labels/10)
            unparentlab <- sort(unique(parentlab))
            pk <- length(unparentlab)
            pss <- NULL
            for (i in seq_len(pk)) {
                if (between == "med")
                    pss[i] <- median(ss[trunc(unlabels/10) == unparentlab[i]],
                                     na.rm = TRUE)
                if (between == "mean")
                    pss[i] <- mean(ss[trunc(unlabels/10) == unparentlab[i]],
                                   na.rm = TRUE)
            }
            if (between == "med")
                out <- median(pss,na.rm = TRUE)
            if (between == "mean")
                out <- mean(pss,na.rm = TRUE)
        }
        else{
            if (between == "med")
                out <- median(ss,na.rm = TRUE)
            if (between == "mean")
                out <- mean(ss,na.rm = TRUE)
        }
    }
    return(out)
}


silcheck <- function(data, kmax=9, diss=FALSE, echo=FALSE, graph=FALSE)
{
    if ( !diss ) {
        if ( inherits(data, "dist"))
            stop("data argument is a dist object, but diss is FALSE")
        if ( is.matrix(data) && (nrow(data)  ==  ncol(data) ) )
            warning("data argument is square, could be a dissimilarity")
    }
    if ( diss && is.matrix(data) && nrow(data) != ncol(data) )
        stop("should be a dissimilarity matrix - but is not square")

    sil <- NULL
    m <- min(kmax, max((!diss)*(dim(data)[1] - 1),
                       (diss)*(0.5*(sqrt(1 + 8*length(data)) - 1)),
                       na.rm = TRUE))
    if (m < 2)
        out <- c(1, NA)
    else{
        for (i in seq_len((m - 1)))
            sil[i] <- pam(data, k = (i + 1), diss = diss)$silinfo$avg.width
        if (echo)
            cat("best k = ", order(sil)[length(sil)] + 1, ", sil(k) = ",
                round(max(sil),4), "\n")
        if (graph) {
            plot(2:m, sil, type = "n", xlab = "Number of Clusters",
                 ylab = "Average Silhouette")
            text(2:m, sil, 2:m)
        }
        out <- c(order(sil)[length(sil)] + 1,max(sil))
    }
    return(out)
}

#msscheck (mss)
msscheck <- function(dist, kmax=9, khigh=9, within="med", between="med",
                     force=FALSE, echo=FALSE, graph=FALSE) {
    p = methods::slot(dist, "Size")
    if (p < 3)
        out <- c(1,NA)
    else{
        dvec <- methods::slot(dist, "Data")

        kmax <- min(kmax,p - 1,na.rm = TRUE)
        if (force)
            mss <- 0
        else
            mss <- labelstomss(rep(1,p),dist,khigh,within,between)
        for (k in 2:kmax)
            mss[k] <- labelstomss(pam(dvec, k, diss = TRUE)$clust, dist,
                                  khigh, within, between)
        shift <- 0
        if (force) {
            mss <- mss[-1]
            shift <- 1
        }
        if (echo)
            cat("best k = ", order(mss)[1] + shift, ", mss(k) = ",
                round(min(mss),4), "\n")
        if (graph) {
            kmin <- ifelse(force,2,1)
            plot(kmin:kmax, mss, type = "n" ,xlab = "Number of Clusters",
                 ylab = "MSS")
            text(kmin:kmax,mss,kmin:kmax)
        }
        out <- c(order(mss)[1] + shift,min(mss))
    }
    return(out)
}

#2. Utility functions

#counts the number of digits in label
digits <- function(label) {
    label <- label[1]
    count <- 0
    while (label >= 1) {
        count <- count + 1
        label <- label/10
    }
    return(count)
}


# patrick's suggestion...
cutdigits <- function(labels, dig) {
    df <- max(0,digits(labels) - dig)
    return(trunc(labels/(10^df)))
}

#removes trailing zeros from labels
cutzeros <- function(labels) {
    for (i in seq_len(length(labels))) {
        while (trunc(labels[i]/10)*10 == labels[i]) {
            labels[i] <- labels[i]/10
        }
    }
    return(labels)
}

#returns the number of non-zero digits in labels
nonzeros <- function(labels) {
    for (i in seq_len(length(labels))) {
        while (trunc(labels[i]/10)*10 == labels[i]) {
            labels[i] <- labels[i]/10
        }
        labels[i] <- digits(labels[i])
    }
    return(labels)
}


msssplitcluster <- function(clust1, l1, id1, medoid1,
                            med2dist, right, dist1,
                            kmax = 9, khigh = 9,
                            within = "med", between = "med") {
    if (!medoid1)
        warning("Medoid missing - continue to split cluster")
    else{
        if (sum(medoid1 == id1) == 0 & medoid1)
            warning("Medoid not in cluster - continue to split cluster")
    }
    if (is.matrix(clust1)) {
        p1 <- length(clust1[,1])
        n <- length(clust1[1,])
    }
    else p1 <- 1
    if (p1 < 3)
        k1 <- 1
    else{
        l <- length(clust1[,1])
        dissvec <- methods::slot(dist1, "Data")

        kmax <- min(p1 - 1,kmax,na.rm = TRUE)
        khigh <- min(p1 - 1,khigh,na.rm = TRUE)
        k1 <- msscheck(dist1, kmax, khigh, within, between)[1]
        if (k1 > 1) {
            pamobj <- pam(dissvec, k1, diss = TRUE)
            newclussizes <- pamobj$clusinfo[,1]
            newmedoids1 <- id1[pamobj$medoids]
            newlabels1 <- pamobj$clustering
            distnewmedoids <- NULL
            for (j in seq_len(k1))
                distnewmedoids[j] <-
                mean(med2dist[newlabels1 == newlabels1[pamobj$medoids[j]]])
            if (right == 1)
                #ord <- order(distnewmedoids, decreasing = TRUE)
                ord <- rev(order(distnewmedoids))
            else
                ord <- order(distnewmedoids)
            newmedoids1 <- newmedoids1[ord]
            newclussizes <- newclussizes[ord]
            oldlab <- newlabels1
            for (j in seq_len(k1))
                newlabels1[oldlab == ord[j]] <- j
            newlabels1 <- rep(10*l1,l) + newlabels1
        }
    }
    if (k1 == 1) {
        newmedoids1 <- medoid1
        newlabels1 <- rep(10*l1,p1)
        newclussizes <- p1
    }
    for (a in seq_len(length(newmedoids1))) {
        if (sum(newmedoids1[a] == id1) == 0)
            warning("Problem with new medoids after splitting cluster")
    }
    list(k1,newmedoids1,newlabels1,newclussizes)
}



mssnextlevel <- function(data ,prevlevel, dmat, kmax = 9,
                         khigh = 9, within = "med", between = "med") {
    n <- length(data[1,])
    p <- length(data[,1])

    id <- seq_len(p)
    k <- prevlevel[[1]]
    medoids <- prevlevel[[2]]
    labels <- prevlevel[[4]]
    newk <- 0
    newlabels <- newmedoids <- newclussizes <- NULL
    count <- 1
    ordlabels <- sort(unique(labels))
    if (length(ordlabels) != k)
        warning("Number of unique labels not equal
                number of clusters in mssnextlevel()")
    if (sum(is.na(medoids))) {
        warning("Missing values in medoid vector in nextlevel()")
        medoids[is.na(medoids)] <- FALSE
    }
    if (length(unique(medoids)) < k && sum(medoids))
        warning("Medoids not unique in mssnextlevel()")
    checkmeans <- FALSE
    if (length(medoids) == 1 && !medoids) {
        warning("No medoids provided in mssnxtlevel()")
        usemean <- TRUE
    }
    else{
        if (sum(medoids > 1) == k)
            usemean <- FALSE
        else
            checkmeans <- TRUE
    }
    for (j in (seq_len(k))) {
        clust1 <- data[labels == ordlabels[j],]
        id1 <- id[labels == ordlabels[j]]
        if (length(id1) > 1)
            clust1 <- as.matrix(clust1)
        l1 <- ordlabels[j]
        right <- (j < k)
        medoid1 <- ifelse(is.na(medoids[j]),0,medoids[j])
        if (j < k)
            medoid2 <- medoids[j + 1]
        else
            medoid2 <- medoids[j - 1]
        if (length(id1) > 1) {
            med2dist <- rowMeans(as(dmat[labels == ordlabels[j],
                                         labels == labels[medoid2],
                                         drop = FALSE],"matrix"))
        }else{
            med2dist <- mean(dmat[labels == ordlabels[j],
                                  labels == labels[medoid2]])
        }

        splitobj <- msssplitcluster(clust1,l1,id1,medoid1,med2dist,right,
                                    dmat[labels == l1,labels == l1],
                                    kmax,khigh,within,between)
        newlabels[labels == ordlabels[j]] <- splitobj[[3]]
        k1 <- splitobj[[1]]
        start <- count
        end <- count + k1 - 1
        newmedoids[start:end] <- splitobj[[2]]
        newclussizes[start:end] <- splitobj[[4]]
        count <- count + k1
    }
    count <- newk <- count - 1
    newmedoids <- newmedoids[seq_len(newk)]
    newclussizes <- newclussizes[seq_len(newk)]
    final <- 0
    if (count == k)
        final <- 1
    if (max(newclussizes) == 3)
        final <- 1
    list(newk,newmedoids,newclussizes,newlabels,final,
         rbind(prevlevel[[6]],cbind(sort(unique(newlabels)),newmedoids)))
}

orderelements  <-  function(level, data, rel = "own",
                            d = "cosangle", dmat = NULL) {

    idn <- seq_len(length(data[,1]))
    k <- level[[1]]
    labels <- level[[4]]
    medoids <- level[[2]]
    clussizes <- level[[3]]
    ord <- order(labels)
    idnord <- idn[ord]
    subdataord <- data[ord,]

    if (is.null(dmat))
        dmat <- distancematrix(data,d)
    distord <- dmat[ord,]

    labelsord <- labels[ord]
    count <- 1
    for (j in (seq_len(k))) {
        start <- count
        end <- count + clussizes[j] - 1
        if (clussizes[j] > 2) {
            tempid <- idnord[start:end]
            if (rel == "co") {
                distj <- distord[,ord][start:end,start:end]
                idnord[start:end] <- tempid[improveordering(distj)]
            }
            else{
                if (rel == "neighbor") {
                    if (j < k)
                        mednext <- medoids[j + 1]
                    else
                        mednext <- medoids[j - 1]
                }else
                    mednext <- medoids[j]

                dmednext <- distord[start:end,mednext]

                if (rel == "neighbor") {
                    if (j < k)
                        #ordtemp <- order(dmednext, decreasing = TRUE)
                        ordtemp <- rev(order(dmednext))
                    else
                        ordtemp <- order(dmednext)
                }
                else
                    ordtemp <- order(dmednext)
                idnord[start:end] <- tempid[ordtemp]
            }
        }
        else
            idnord[start:end] <- idnord[start:end]
        count <- count + clussizes[j]
    }
    list(data[idnord,],idnord)
}



mssinitlevel <- function(data, kmax=9, khigh=9, d="cosangle", dmat=NULL,
                         within="med", between="med", ord="co",
                         verbose=FALSE)
{
    #print("mssinitlevel")
    if (!is.matrix(data))
        stop("First arg to mssinitlevel() must be a matrix")

    p <- length(data[,1])

    if (is.matrix(dmat))
        dmat <- as.hdist(dmat)

    else
        dmat <- distancematrix(data,d = d)

    if (methods::slot(dmat, "Size") != p)
        stop("Data and distance matrix dimensions
             do not agree in mssinitlevel()")

    m <- msscheck(dmat,kmax,khigh,within,between)
    if (m[1] == 1) {
        if (verbose)
            cat("No strong evidence for clusters in the first level -
          \n continuing to split root node anyway. \n")
        m <- msscheck(dmat,kmax,khigh,within,between,force = TRUE)
    }

    pamobj <- pam(methods::slot(dmat, "Data"), m[1], diss = TRUE)
    rowmedoids <- pamobj$medoids
    final <- ifelse(max(pamobj$clusinfo[,1]) < 3,1,0)
    if (m[1] > 2) {
        medoidsdata <- as.matrix(data[rowmedoids,])
        l <- length(rowmedoids)
        medoidsdist <- dmat[rowmedoids,rowmedoids]
        if (ord == "co")
            medoidsord <- improveordering(medoidsdist)
        if (ord == "clust") {
            mpamobj <- pam(methods::slot(medoidsdist, "Size"), 2, diss = TRUE)
            labelsmed <- mpamobj$clustering
            medmed <- mpamobj$medoids
            clussizes <- mpamobj$clusinfo[,1]

            prevlevel <- mssnextlevel(medoidsdata,list(2,medmed,clussizes,
                                                       labelsmed,0,
                                                       cbind(c(1,2),medmed)),
                                      dmat = medoidsdist,kmax,khigh,within,
                                      between)
            final <- prevlevel[[5]]
            if (final == 0) {
                depth <- (l - 1)
                for (j in seq_len(depth)) {
                    if (final == 0) {
                        clustnext <- mssnextlevel(medoidsdata,prevlevel,
                                                  dmat = medoidsdist,kmax,
                                                  khigh,within,between)
                        final <- clustnext[[5]]
                    }
                    if (final == 1) {
                        j <- depth
                        prevlevel <- clustnext
                    }
                }
            }
            medoidsord <- orderelements(prevlevel,medoidsdata,rel = "neighbor",
                                        d = d,dmat = medoidsdist)[[2]]
        }
        k <- m[1]
        rowmedoids <- rowmedoids[medoidsord]
        labels <- lab2 <- pamobj$clustering
        for (j in (seq_len(k)))
            lab2[labels == medoidsord[j]] <- j
        output <- list(k,rowmedoids,pamobj$clusinfo[,1][medoidsord],lab2,
                       final,cbind(seq_len(k),rowmedoids))
    }
    else
        output <- list(2,pamobj$medoids,pamobj$clusinfo[,1],
                       pamobj$clustering,final,cbind(seq_len(2),pamobj$medoids))
    return(output)
}


paircoll <- function(i,j,data,level,d="cosangle",
                     dmat=NULL, newmed="medsil") {
    p <- length(data[,1])
    k <- level[[1]]
    labels <- level[[4]]
    medoids <- level[[2]]
    clussizes <- level[[3]]
    N <- length(level[[6]][,1])
    block <- level[[6]][(N - k + 1):N,]
    if (N == k)
        prevblock <- NULL
    else
        prevblock <- level[[6]][seq_len(N - k),]
    if (max(i,j) > k)
        stop("The clusters to collapse do not exist in paircoll()")
    labeli <- labels[medoids[i]]
    labelj <- labels[medoids[j]]
    oldlabels <- labels
    labels[labels == labelj] <- labeli
    trunclabels <- trunc(oldlabels/10)
    labelparents <- unique(trunclabels)
    parenti <- order(labelparents)[labelparents == trunc(labeli/10)]
    parentj <- order(labelparents)[labelparents == trunc(labelj/10)]
    if (newmed == "nn")
        fakemed <- (data[medoids[i],]*clussizes[i] +
                        data[medoids[j],]*clussizes[j])/(clussizes[i] +
                                                             clussizes[j])
    if (newmed == "uwnn")
        fakemed <- (data[medoids[i],] + data[medoids[j],])/2
    if (newmed == "nn" || newmed == "uwnn") {
        rowsub <- (seq_len(p))[labels == labeli]
        distsfm <- distancevector(data[rowsub,],as.vector(fakemed),d)
        medoids[i] <- rowsub[order(distsfm)[1]]
    }
    else{
        #colldist <- as.matrix(dmat[labels == labeli,labels == labeli])
        colldist <- as(dmat[labels == labeli,labels == labeli],"matrix")

        rowsub <- (seq_len(p))[labels == labeli]
        if (newmed == "center") {
            sumdist <- rowSums(colldist)
        }
        if (newmed == "medsil") {
            othermed <- medoids[-c(i,j)]
            collp <- length(labels[labels == labeli])
            othern <- length(othermed)
            if (othern == 0)
                stop("Not enough medoids to use newmed='medsil' in paircoll()")
            if (is(dmat,"hdist")) {
                if (othern == 1) {
                    #otherdist <- cbind(dmat[labels == labeli,othermed])
                    otherdist <- dmat[labels == labeli,othermed]
                }else{
                    #otherdist <- rbind(dmat[labels == labeli,othermed])
                    otherdist <- dmat[labels == labeli,othermed]
                }
            }

            if (othern == 1)
                b <- otherdist
            else
                b <- apply(otherdist,1,min)
            b <- matrix(b,nrow=collp,ncol=collp)
            diag(b) <- 0
            b <- abs(b-colldist)/pmax(colldist,b)
            sumdist <- rowSums(b)
            medoids[i] <- rowsub[order(sumdist,decreasing=TRUE) == 1]
            #medoids[i] <- rowsub[rev(order(sumdist)) == 1]
        }
    }
    k <- k-1
    clussizes[i] <- clussizes[i] + clussizes[j]
    block[i,2] <- medoids[i]
    if (j<=k) {
        for (l in (j:k)) {
            medoids[l] <- medoids[l + 1]
            clussizes[l] <- clussizes[l + 1]
            block[l,] <- block[l + 1,]
        }
    }
    medoids <- medoids[seq_len(k)]
    clussizes <- clussizes[seq_len(k)]
    block <- block[seq_len(k),]
    if (labels[labels == labeli][1]/10 == trunclabels[labels == labeli][1]) {
        labels[labels == labeli] <- labels[labels == labeli] + 1
        labels[labels == labelj] <- labels[labels == labelj] + 1
        block[i,1] <- block[i,1] + 1
    }
    return(list(k,medoids,clussizes,labels,level[[5]],rbind(prevblock,block)))
}


collap <- function(data,level,d = "cosangle", dmat = NULL, newmed = "medsil") {
    k <- level[[1]]
    if (k < 3) {
        warning("Not enough medoids to use newmed='medsil' in collap() - \n
            using newmed='nn' instead \n")
        newmed <- "nn"
    }
    medoids <- level[[2]]
    clussizes <- level[[3]]
    if (sum(is.na(clussizes)))
        warning("NA in clussizes")
    medoidsdata <- data[medoids,]
    if (sum(is.na(medoidsdata)) > 0)
        warning("Missing value(s) in medoidsdata in collap()")

    distmed <- dmat[medoids,medoids]
    distv <- methods::slot(distmed, "Data")

    indexmin <- order(distv)[1]
    best <- vectmatrix(indexmin,k)
    clustfinal <- paircoll(best[1],best[2],data,level,d,dmat,newmed)
    return(clustfinal)
}

msscollap <- function(data,level,khigh,d="cosangle", dmat = NULL,
                      newmed = "medsil", within = "med",
                      between = "med", impr = 0) {

    newk <- level[[1]]
    mss1 <- labelstomss(level[[4]],dmat,khigh,within,between)
    maxncoll <- max(0,newk - 2)
    ncoll <- 0
    coll <- 1
    if (newk <= 2)
        coll <- 0
    while ((coll == 1) && (ncoll <= maxncoll)) {
        levelc <- collap(data,level,d,dmat,newmed)
        mss2 <- labelstomss(levelc[[4]],dmat,khigh,within,between)
        if (mss1 == 0)
            r <- 0
        else
            r <- (mss1 - mss2)/mss1
        if (r < impr)
            coll <- 0
        else{
            mss1 <- mss2
            level <- levelc
            ncoll <- ncoll + 1
        }
    }
    return(level)
}

mssmulticollap <- function(data,level,khigh,d="cosangle",
                           dmat = NULL, newmed = "medsil",
                           within = "med", between = "med", impr = 0) {
    if (!is.matrix(data))
        stop("First arg to mssmulticollap() must be a matrix")

    medoids <- level[[2]]
    medoidsdata <- data[medoids,]
    if (sum(is.na(medoidsdata)) > 0)
        warning("Missing value(s) in medoidsdata in mssmulticollap()")

    distmed <- dmat[medoids,medoids]
    k <- level[[1]]
    ord <- order(methods::slot(distmed, "Data"))

    mss1 <- labelstomss(level[[4]],dmat,khigh,within,between)
    maxncoll <- max(0,k*(k - 1)/2)
    ncoll <- 0
    i <- 1
    while (i <= maxncoll) {
        clusts <- vectmatrix(ord[i],k)
        if (k < 3) {
            if (newmed == "medsil")
                warning("Can't use newmed=medsil with less than 3 clusters. \n
                        Substituting newmed=nn")
            newmed <- "nn"
        }
        levelc <- paircoll(clusts[1],clusts[2],data,level,d,dmat,newmed)
        mss2 <- labelstomss(levelc[[4]],dmat,khigh,within,between)
        if (mss1 == 0)
            r <- 0
        else
            r <- (mss1 - mss2)/mss1
        if (r >= impr) {
            mss1 <- mss2
            level <- levelc
            ncoll <- ncoll + 1
            k <- level[[1]]
            maxncoll <- max(0,k*(k - 1)/2)
            i <- 0
            medoids <- level[[2]]
            medoidsdata <- data[medoids,]
            if (sum(is.na(medoidsdata)) > 0)
                warning("Missing value(s) in medoidsdata in mssmulticollap()")

            distmed <- dmat[medoids,medoids]
            ord <- order(methods::slot(distmed, "Data"))
        }
        i <- i + 1
    }
    return(level)
}


mssrundown <- function(data, K=16, kmax=9, khigh=9, d="cosangle",
                       dmat=NULL, initord="co", coll="seq",
                       newmed="medsil", stop=TRUE,
                       finish=FALSE, within="med",
                       between="med",impr=0, verbose=FALSE)
{
    #print("mssrundown")
    if (!is.matrix(data))
        stop("First arg to mssrundown() must be a matrix")

    bestlevel <- level <- mssinitlevel(data, kmax, khigh,
                                       d, dmat, within,
                                       between, initord, verbose)
    bestmss <- mss <- labelstomss(level[[4]],dmat,khigh,within,between)
    bestl <- l <- 1
    ind <- 0
    if (verbose)
        cat("Searching for main clusters... \n")
    if (level[[5]] == 1)
        return(level)
    while((l<=K) && (ind == 0)) {
        if (verbose) cat("Level ",l,"\n")
        if (coll == "seq")
            levelc <- msscollap(data,level,khigh,d,dmat,
                                newmed,within,between,impr)
        if (coll == "all")
            levelc <- mssmulticollap(data,level,khigh,d,
                                     dmat,newmed,within,between,impr)
        mss <- labelstomss(levelc[[4]],dmat,khigh,within,between)
        if (mss>=bestmss & stop == TRUE)
            ind <- 1
        else{
            if (mss<bestmss) {
                bestlevel <- levelc
                bestmss <- mss
                bestl <- l
            }
        }
        l <- l+1
        if (l<=K) {
            level <- mssnextlevel(data,levelc,dmat,kmax,khigh,within,between)
            if (finish == TRUE) {
                if (sum(trunc(level[[4]]/10)*10 == level[[4]]) ==
                    length(level[[4]]) & l<=K) {
                    ind <- 1
                    bestlevel <- levelc
                    bestmss <- mss
                    bestl <- (l-1)
                }
            }
        }
    }
    if (verbose)
        cat("Identified", bestlevel[[1]],
            " main clusters in level",
            bestl, "with MSS =",bestmss,"\n")
    return(bestlevel)
}

msscomplete <- function(level, data, K=16, khigh=9, d="cosangle",
                        dmat=NULL, within="med", between="med", verbose=FALSE)
{
    if (!is.matrix(data))
        stop("First arg to msscomplete() must be a matrix")

    count <- digits(level[[4]][1])
    if (verbose)
        cat("Running down without collapsing from Level",count,"\n")
    while ((max(level[[3]]) > 3) & (count < K)) {
        level <- newnextlevel(data,level,dmat,2,khigh)
        count <- count + 1
        if (verbose) cat("Level",count,"\n")
    }
    return(level)
}



newnextlevel <- function(data, prevlevel, dmat, klow=2, khigh=6) {
    #print("newnextlevel")
    if (!is.matrix(data))
        stop("First arg to newnextlevel() must be a matrix")
    p <- length(data[,1])
    n <- length(data[1,])
    id <- seq_len(p)
    k <- prevlevel[[1]]
    medoids <- prevlevel[[2]]
    labels <- prevlevel[[4]]
    newk <- 0
    newlabels <- newmedoids <- newclussizes <- NULL
    count <- 1
    ordlabels <- sort(unique(labels))
    if (length(ordlabels) != k)
        warning("Number of unique labels not equal
                number of clusters in newnextlevel()")
    if (sum(is.na(medoids))) {
        warning("Missing value(s) in medoid vector in newnextlevel()")
        medoids[is.na(medoids)] <- FALSE
    }
    if (length(unique(medoids)) < k && sum(medoids))
        warning("Medoids in newnextlevel() are not unique")
    checkmeans <- FALSE
    if (length(medoids) == 1 && !medoids) {
        warning("No medoids provided in newnextlevel()")
        usemean <- TRUE
    }
    else{
        if (sum(medoids > 1) == k)
            usemean <- FALSE
        else
            checkmeans <- TRUE
    }
    for (j in (seq_len(k))) {
        clust1 <- data[labels == ordlabels[j],]
        id1 <- id[labels == ordlabels[j]]
        if (length(id1) > 1) {
            kmax <- min(c(khigh,dim(clust1)[1] - 1))
            clust1 <- as.matrix(clust1)
        }
        l1 <- ordlabels[j]
        right <- (j < k)
        medoid1 <- ifelse(is.na(medoids[j]),0,medoids[j])
        if (j < k)
            medoid2 <- medoids[j + 1]
        else
            medoid2 <- medoids[j - 1]
        if (length(id1) > 1)
            med2dist <- rowMeans(as(dmat[labels == ordlabels[j],
                                         labels == labels[medoid2],
                                         drop = FALSE],"matrix"))
        else
            med2dist <- mean(dmat[labels == ordlabels[j],
                                  labels == labels[medoid2]])

        splitobj <- newsplitcluster(clust1,l1,id1,klow,kmax,medoid1,med2dist,
                                    right,dmat[labels == l1,labels == l1])
        newlabels[labels == ordlabels[j]] <- splitobj[[3]]
        k1 <- splitobj[[1]]
        start <- count
        end <- count + k1 - 1
        newmedoids[start:end] <- splitobj[[2]]
        newclussizes[start:end] <- splitobj[[4]]
        count <- count + k1 - 1 + 1
    }
    count <- count - 1
    newk <- count
    newmedoids <- newmedoids[seq_len(newk)]
    newclussizes <- newclussizes[seq_len(newk)]
    final <- 0
    if (count == k)
        final <- 1
    if (max(newclussizes) == 3)
        final <- 1
    return(list(newk,newmedoids,newclussizes,newlabels,
                final,rbind(rbind(prevlevel[[6]],
                                  cbind(sort(unique(newlabels)),newmedoids)))))
}

newsplitcluster <- function(clust1, l1, id1, klow=2, khigh=2, medoid1,
                            med2dist, right, dist1) {
    #print("newsplitcluter")
    if (!medoid1)
        warning("Medoid missing - continue to split cluster")
    else{
        if (sum(medoid1 == id1) == 0 & medoid1)
            warning("Medoid not in cluster - continue to split cluster")
    }
    if (is.matrix(clust1)) {
        p1 <- length(clust1[,1])
        n <- length(clust1[1,])
    }
    else p1 <- 1
    if (p1 < 3) {
        k1 <- 1
        newmedoids1 <- medoid1
        newlabels1 <- rep(10*l1,p1)
        newclussizes <- p1
    }
    else{
        l <- length(clust1[,1])
        dissvec <- methods::slot(dist1, "Data")

        kmax <- min(p1 - 1, khigh)
        a <- rep(0,(kmax - klow + 2))
        best <- 2
        for (j in (klow:kmax)) {
            a[j] <- pam(dissvec,j,diss = TRUE)$silinfo$avg.width
            if (a[j] > a[best]) best <- j
        }
        k1 <- best
        pamobj <- pam(dissvec, k1, diss = TRUE)
        newclussizes <- pamobj$clusinfo[,1]
        newmedoids1 <- id1[pamobj$medoids]
        newlabels1 <- pamobj$clustering
        distnewmedoids <- NULL
        for (j in seq_len(k1))
            distnewmedoids[j] <-
            mean(med2dist[newlabels1 ==
                              newlabels1[pamobj$medoids[j]]])
        if (right == 1)
            ord <- order(distnewmedoids, decreasing = TRUE)
        #ord <- rev(order(distnewmedoids))
        else
            ord <- order(distnewmedoids)
        newmedoids1 <- newmedoids1[ord]
        newclussizes <- newclussizes[ord]
        oldlab <- newlabels1
        for (j in seq_len(k1))
            newlabels1[oldlab == ord[j]] <- j
        newlabels1 <- rep(10*l1,l) + newlabels1
    }
    for (a in seq_len(length(newmedoids1))) {
        if (sum(newmedoids1[a] == id1) == 0)
            warning("Problem with new medoids after splitting cluster")
    }
    return(list(k1,newmedoids1,newlabels1,newclussizes))
}



hopach <- function(data, dmat=NULL, d="cosangle", clusters="best", K=15,
                   kmax=9, khigh=9, coll="seq", newmed="medsil",
                   mss="med", impr=0, initord="co", ord="own", verbose=FALSE) {
    # if (inherits(data,"ExpressionSet"))
    #   data  <-  exprs(data)
    data <- as.matrix(data)

    # Convert to hdist immediately #
    if ( is.null(dmat) ) {
        dmat <- distancematrix(data, d)
    }else if ( is.matrix(dmat) ) {
        dmat <- as(dmat, "hdist")
    }else if ( "dist" %in% is(dmat)) {
        dmat <- hdist(Data = as.numeric(dmat), Size = attr(dmat, "Size"),
                      Labels = seq_len(attr(dmat, "Size")),
                      Call = as.character(attr(dmat, "call"))[3])
    }else if (!is.hdist(dmat)) {
        stop("Distance matrix could not be created into hdist object.")
    }

    if (K > 15) {
        K <- 15
        warning("K set to 15 - can't do more than 15 splits")
    }
    if (K < 1) {
        K <- 1
        warning("K set to 1 - can't do less than 1 level")
    }
    if (clusters != "none") {
        cuttree <- mssrundown(data, K, kmax, khigh, d, dmat,
                              initord, coll, newmed,
                              stop = (clusters == "greedy"), finish = TRUE,
                              within = mss,
                              between = mss, impr, verbose)
        if (cuttree[[1]] > 1)
            cutord <- orderelements(cuttree, data, rel = ord, d, dmat)[[2]]
        else
            cutord <- NULL
        out1 <- list(k = cuttree[[1]], medoids = cuttree[[2]],
                     sizes = cuttree[[3]], labels = cuttree[[4]],
                     order = cutord)
        finaltree <- msscomplete(cuttree, data, K, khigh, d,
                                 dmat, within = mss, between = mss,
                                 verbose)
    }
    else{
        out1 <- NULL
        finaltree <- msscomplete(mssinitlevel(as.matrix(data),
                                              kmax, khigh, d, dmat,
                                              within = mss, between = mss,
                                              initord), data, K, khigh,
                                 d, dmat, within = mss,
                                 between = mss, verbose)
    }
    dimnames(finaltree[[6]]) <- list(NULL,c("label","medoid"))
    out2 <- list(labels = finaltree[[4]],
                 order = orderelements(finaltree, data, rel = ord, d,
                                       dmat)[[2]], medoids = finaltree[[6]])
    return(list(clustering = out1, final = out2, call = match.call(),
                metric = d))
}

distancematrix  <-  function(X, d, na.rm = TRUE)
{
    X  <-  as.matrix(X)
    if (d == "euclid")
        return(hopach::disseuclid(X, na.rm))
    if (d == "cor")
        return(hopach::disscor(X, na.rm))
    if (d == "abscor")
        return(hopach::dissabscor(X, na.rm))
    if (d == "cosangle")
        return(hopach::disscosangle(X, na.rm))
    if (d == "abscosangle")
        return(hopach::dissabscosangle(X, na.rm))
    stop("Distance metric ", d, " not available")
}

#' Testing scClassify model
#'
#' @param exprsMat_test A list or a matrix indicates the log-transformed
#' expression matrices of the query datasets
#' @param trainRes A `scClassifyTrainModel` or a `list` indicates
#' scClassify trained model
#' @param cellTypes_test A list or a vector indicates cell types
#' of the qurey datasets (Optional).
#' @param k An integer indicates the number of neighbour
#' @param prob_threshold A numeric indicates the probability threshold
#' for KNN/WKNN/DWKNN.
#' @param cor_threshold_static A numeric indicates the static
#' correlation threshold.
#' @param cor_threshold_high A numeric indicates the highest
#' correlation threshold
#' @param features A vector indicates the gene selection method,
#' set as "limma" by default.
#' This should be one or more of "limma", "DV", "DD", "chisq", "BI".
#' @param algorithm A vector indicates the KNN method that are used,
#' set as "WKNN" by default.
#' This should be one or more of "WKNN", "KNN", "DWKNN".
#' @param similarity A vector indicates the similarity measure that are used,
#' set as "pearson" by default.
#' This should be one or more of "pearson",  "spearman", "cosine",
#' "jaccard", "kendall", "binomial", "weighted_rank","manhattan"
#' @param cutoff_method A vector indicates the method to cutoff the
#' correlation distribution. Set as "dynamic" by default.
#' @param weighted_ensemble A logical input indicates in ensemble learning,
#' whether the results is combined by a
#' weighted score for each base classifier.
#' @param weights A vector indicates the weights for ensemble
#' @param parallel A logical input indicates whether running in paralllel or not
#' @param BPPARAM  A \code{BiocParallelParam} class object
#' from the \code{BiocParallel} package is used. Default is SerialParam().
#' @param verbose A logical input indicates whether the intermediate steps
#' will be printed
#' @return list of results
#'
#' @examples
#' data("scClassify_example")
#' wang_cellTypes <- scClassify_example$wang_cellTypes
#' exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
#' data("trainClassExample_xin")
#'
#' pred_res <- predict_scClassify(exprsMat_test = exprsMat_wang_subset,
#' trainRes = trainClassExample_xin,
#' cellTypes_test = wang_cellTypes,
#' algorithm = "WKNN",
#' features = c("limma"),
#' similarity = c("pearson"),
#' prob_threshold = 0.7,
#' verbose = TRUE)
#'
#' @author Yingxin Lin
#'
#' @importFrom methods is
#' @importFrom BiocParallel SerialParam bplapply
#'
#' @export

# function to predict multiple feature + distance combinations

predict_scClassify <- function(exprsMat_test,
                               trainRes,
                               cellTypes_test = NULL,
                               k = 10,
                               prob_threshold = 0.7,
                               cor_threshold_static = 0.5,
                               cor_threshold_high = 0.7,
                               features = "limma",
                               algorithm = "WKNN",
                               similarity = "pearson",
                               cutoff_method = c("dynamic", "static"),
                               weighted_ensemble = FALSE,
                               weights = NULL,
                               parallel = FALSE,
                               BPPARAM = BiocParallel::SerialParam(),
                               verbose = FALSE){

    # checking input

    algorithm <- match.arg(algorithm, c("WKNN", "KNN", "DWKNN"),
                           several.ok = TRUE)
    similarity <- match.arg(similarity, c("pearson",  "spearman",
                                          "cosine", "jaccard", "kendall",
                                          "weighted_rank","manhattan"),
                            several.ok = TRUE)
    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"),
                               several.ok = TRUE)


    if ("scClassifyTrainModelList" %in% is(trainRes)) {
        stop("For a list of training model,
         please use predict_scClassifyJoint()
             instead to get joint training results.")
    }

    if (!any(c("scClassifyTrainModel", "list") %in% is(trainRes))) {
        stop("Wrong trainRes input.
             Need to be either scClassifyTrainModel or list")
    }

    if ((length(features) > 1 | length(algorithm) > 1 |
         length(similarity) > 1) ) {
        if (weighted_ensemble) {
            weighted_ensemble <- TRUE
            ensemble <- TRUE

            if (verbose) {
                cat("Performing weighted ensemble learning... \n")
            }

        } else {
            weighted_ensemble <- FALSE
            ensemble <- TRUE

            if (verbose) {
                cat("Performing unweighted ensemble learning... \n")
            }

        }
    } else {
        weighted_ensemble <- FALSE
        ensemble <- FALSE

        if (verbose) {
            cat("Ensemble learning is disabled... \n")
        }

    }



    ensemble_methods <- as.matrix(expand.grid(similarity = similarity,
                                              algorithm = algorithm,
                                              features = features))

    if (!is.null(weights)) {
        if (length(weights) != nrow(ensemble_methods)) {
            stop("The length of weights is not equal to
           the number of combination of ensemble methods")
        }
    }



    if (parallel) {

        predictRes <- BiocParallel::bplapply(seq_len(nrow(ensemble_methods)),
                                             function(em){
                                                 predict_scClassifySingle(exprsMat_test =  exprsMat_test,
                                                                          trainRes = trainRes,
                                                                          cellTypes_test = cellTypes_test,
                                                                          k = k,
                                                                          prob_threshold = prob_threshold,
                                                                          cor_threshold_static = cor_threshold_static,
                                                                          cor_threshold_high = cor_threshold_high,
                                                                          features = ensemble_methods[em, 3],
                                                                          algorithm = ensemble_methods[em, 2],
                                                                          similarity = ensemble_methods[em, 1],
                                                                          cutoff_method = cutoff_method,
                                                                          verbose = verbose)
                                             }, BPPARAM = BPPARAM)


    }else{
        predictRes <- list()
        for (em in seq(nrow(ensemble_methods))) {

            if (verbose) {
                cat("Using parameters: \n")
                print(ensemble_methods[em, ])
            }

            predictRes[[em]] <- predict_scClassifySingle(exprsMat_test =  exprsMat_test,
                                                         trainRes = trainRes,
                                                         cellTypes_test = cellTypes_test,
                                                         k = k,
                                                         prob_threshold = prob_threshold,
                                                         cor_threshold_static = cor_threshold_static,
                                                         cor_threshold_high = cor_threshold_high,
                                                         features = ensemble_methods[em, 3],
                                                         algorithm = ensemble_methods[em, 2],
                                                         similarity = ensemble_methods[em, 1],
                                                         cutoff_method = cutoff_method,
                                                         verbose = verbose)
        }


    }
    names(predictRes) <- paste(ensemble_methods[, 1],
                               ensemble_methods[, 2],
                               ensemble_methods[, 3],
                               sep = "_")

    if (is.null(weights)) {
        if ("list" %in% is(trainRes)) {
            weights <- trainRes$modelweights
            weights <- weights[names(predictRes)]
        }

        if ("scClassifyTrainModel" %in% is(trainRes)) {
            weights <- modelweights(trainRes)
            weights <- weights[names(predictRes)]
        }

    }

    
    if (verbose) {
        cat("weights for each base method: \n")
        print(weights)
    }



    if (ensemble) {
        ensembleRes <- getEnsembleRes(predictRes,
                                      weights,
                                      exclude = NULL,
                                      weighted_ensemble = weighted_ensemble)


        predictRes$ensembleRes <- ensembleRes
    }


    return(predictRes)
}


#' Testing scClassify model (joint training)
#'
#' @param exprsMat_test A list or a matrix indicates the expression matrices of the testing datasets
#' @param trainRes A `scClassifyTrainModel` or a `list` indicates scClassify training model
#' @param cellTypes_test A list or a vector indicates cell types of the testing datasets (Optional).
#' @param k An integer indicates the number of neighbour
#' @param prob_threshold A numeric indicates the probability threshold for KNN/WKNN/DWKNN.
#' @param cor_threshold_static A numeric indicates the static correlation threshold.
#' @param cor_threshold_high A numeric indicates the highest correlation threshold
#' @param features A vector indicates the method to select features, set as "limma" by default.
#' This should be one or more of "limma", "DV", "DD", "chisq", "BI".
#' @param algorithm A vector indicates the KNN method that are used, set as "WKNN" by default.
#' This should be one or more of "WKNN", "KNN", "DWKNN".
#' @param similarity A vector indicates the similarity measure that are used,
#' set as "pearson" by default.
#' This should be one or more of "pearson",  "spearman", "cosine", "jaccard", "kendall",
#' "binomial", "weighted_rank","manhattan"
#' @param cutoff_method A vector indicates the method to cutoff the correlation distribution.
#' Set as "dynamic" by default.
#' @param parallel A logical input indicates whether running in paralllel or not
#' @param BPPARAM  A \code{BiocParallelParam} class object
#' from the \code{BiocParallel} package is used. Default is SerialParam().
#' @param verbose A logical input indicates whether the intermediate steps will be printed
#' @return list of results
#' @author Yingxin Lin
#'
#' @examples
#' data("scClassify_example")
#' wang_cellTypes <- scClassify_example$wang_cellTypes
#' exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
#' data("trainClassExample_xin")
#' data("trainClassExample_wang")
#'
#' trainClassExampleJoint <- scClassifyTrainModelList(trainClassExample_wang,
#' trainClassExample_xin)
#'
#' pred_res_joint <- predict_scClassifyJoint(exprsMat_test = exprsMat_wang_subset,
#' trainRes = trainClassExampleJoint,
#' cellTypes_test = wang_cellTypes,
#' algorithm = "WKNN",
#' features = c("limma"),
#' similarity = c("pearson"),
#' prob_threshold = 0.7,
#' verbose = FALSE)
#'
#' table(pred_res_joint$jointRes$cellTypes, wang_cellTypes)
#'
#' @importFrom methods is slot
#' @importFrom BiocParallel SerialParam
#' @export

predict_scClassifyJoint <- function(exprsMat_test,
                                    trainRes,
                                    cellTypes_test = NULL,
                                    k = 10,
                                    prob_threshold = 0.7,
                                    cor_threshold_static = 0.5,
                                    cor_threshold_high = 0.7,
                                    features = "limma",
                                    algorithm = "WKNN",
                                    similarity = "pearson",
                                    cutoff_method = c("dynamic", "static"),
                                    parallel = FALSE,
                                    BPPARAM = BiocParallel::SerialParam(),
                                    verbose = FALSE){

    # checking input

    algorithm <- match.arg(algorithm, c("WKNN", "KNN", "DWKNN"),
                           several.ok = TRUE)
    similarity <- match.arg(similarity, c("pearson",  "spearman",
                                          "cosine", "jaccard", "kendall",
                                          "weighted_rank","manhattan"),
                            several.ok = TRUE)
    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"),
                               several.ok = TRUE)


    if (!"scClassifyTrainModelList" %in% is(trainRes)) {
        stop("trainRes needs to be a scClassifyTrainModelList object")
    }



    predictResList <- list()


    for (trainListIdx in seq_len(length(methods::slot(trainRes, "listData")))) {

        if (verbose) {
            cat("Training using ")
            cat(names(methods::slot(trainRes, "listData"))[trainListIdx], "\n")
        }

        predictResList[[trainListIdx]] <- predict_scClassify(exprsMat_test = exprsMat_test,
                                                             trainRes = trainRes[[trainListIdx]],
                                                             cellTypes_test = cellTypes_test,
                                                             k = k,
                                                             prob_threshold = prob_threshold,
                                                             cor_threshold_static = cor_threshold_static,
                                                             cor_threshold_high = cor_threshold_high,
                                                             features = features,
                                                             algorithm = algorithm,
                                                             similarity = similarity,
                                                             cutoff_method = cutoff_method,
                                                             parallel = parallel,
                                                             BPPARAM = BPPARAM,
                                                             verbose = verbose)
    }


    names(predictResList) <- paste("Trained_by",
                                   names(methods::slot(trainRes, "listData")),
                                   sep = "_")

    jointRes <- getJointRes(predictResList, trainRes)

    rownames(jointRes) <- colnames(exprsMat_test)


    predictResList$jointRes <- jointRes


    return(predictResList)
}


predict_scClassifySingle <- function(exprsMat_test,
                                     trainRes,
                                     cellTypes_test,
                                     k = 10,
                                     prob_threshold = 0.7,
                                     cor_threshold_static = 0.5,
                                     cor_threshold_high = 0.7,
                                     features = "limma",
                                     algorithm = c("WKNN", "KNN", "DWKNN"),
                                     similarity = c("pearson",  "spearman",
                                                    "cosine", "jaccard",
                                                    "kendall", "weighted_rank",
                                                    "manhattan"),
                                     cutoff_method = c("dynamic", "static"),
                                     verbose = TRUE){


    if (!is.null(cellTypes_test)) {
        if (length(cellTypes_test) != ncol(exprsMat_test)) {
            stop("Length of testing cell types does not match
           with number of column of testing expression matrix")
        }
    }

    if (is.null(rownames(exprsMat_test))) {
        stop("rownames of exprsMat_test is NULL")
    }

    if (is.null(colnames(exprsMat_test)) |
        sum(duplicated(colnames(exprsMat_test))) != 0) {
        stop("colnames of exprsMat_test is NULL or not unique")
    }

    if (all(exprsMat_test %% 1 == 0)) {
        warning("exprsMat_test looks like a count matrix
            (scClassify requires a log-transformed normalised input)")
    }

    # check input
    algorithm <- match.arg(algorithm, c("WKNN", "KNN", "DWKNN"))
    similarity <- match.arg(similarity, c("pearson",  "spearman",
                                          "cosine", "jaccard", "kendall",
                                          "weighted_rank","manhattan"))
    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"))


    if (ncol(exprsMat_test) < 100 & cutoff_method == "dynamic") {
        warning("Number of cells in test data is small.
                The cutoff method is changed to static.")
        cutoff_method <- "static"
    }


    if ("scClassifyTrainModelList" %in% is(trainRes)) {
        stop("For a list of training model,
         please use predict_scClassifyJoint instead to
             get joint training results.")
    }

    if (!any(c("scClassifyTrainModel", "list") %in% is(trainRes))) {
        stop("wrong trainRes input.
             Need to be either scClassifyTrainModel or list")
    }

    if ("scClassifyTrainModel" %in% is(trainRes)) {

        if (!features %in% features(trainRes)) {
            stop("The selected features are not trained in the provided model!")
        }


        # get the input from train model

        levelModel <- model(trainRes)[[features]]$model
        levelHVG <- model(trainRes)[[features]]$hvg
        cutree_list <- cellTypeTree(trainRes)
        cellTypes_train <- cellTypeTrain(trainRes)
    } else {

        if (!features %in% names(trainRes$hierarchyKNNRes)) {
            stop("The selected features are not trained in the provided model!")
        }


        # get the input from train model

        levelModel <- trainRes$hierarchyKNNRes[[features]]$model
        levelHVG <- trainRes$hierarchyKNNRes[[features]]$hvg
        cutree_list <- trainRes$cutree_list
        cellTypes_train <- trainRes$cellTypes_train

    }

    pred <- list()

    # For each level
    for (i in seq_len(length(levelModel))) {
        # If this level is not NULL (Not Level1)
        if (!is.null(levelModel[[i]])) {


            #If not all the cells are not unassigned in the parent nodes
            if (sum(pred[[i - 1]] != 0) != 0) {
                pred_level <- list()

                for (j in seq_len(length(levelModel[[i]]))) {

                    # If the model of level i-1, cells labeled as j is NOT "no model"
                    if (!"character" %in% is(levelModel[[i]][[j]])) {

                        # Select the cells that are going to classified
                        # (according to what they are classified in last level)
                        predIdx <- which(pred[[i - 1]] == j)

                        if (length(predIdx) != 0) {

                            # features that are in the test dataset
                            common_HVG <- intersect(rownames(exprsMat_test), levelHVG[[i]][[j]])
                            exprsMat_toTest <- exprsMat_test[common_HVG, predIdx, drop = FALSE]
                            exprsMat_toTest <- exprsMat_toTest[Matrix::rowSums(exprsMat_toTest) != 0, ,
                                                               drop = FALSE]

                            if (nrow(exprsMat_toTest) < 5) {

                                message(paste("There are only", nrow(exprsMat_toTest),
                                              "selected genes in reference data expressed in query data"))
                                pred_level[[j]] <- rep(0, length(predIdx))
                            }else{

                                # Calculate the similarity
                                corMat <- calculateSimilarity(exprsMat_train = Matrix::t(levelModel[[i]][[j]]$train[ ,rownames(exprsMat_toTest)]),
                                                              exprsMat_test = exprsMat_toTest,
                                                              similarity = similarity)

                                # Different algorithm
                                if (algorithm == "KNN") {
                                    predRes <- KNNcor(corMat = corMat,
                                                      subLevelModel = levelModel[[i]][[j]],
                                                      cutoff_method = cutoff_method,
                                                      k = k,
                                                      prob_threshold = prob_threshold,
                                                      cor_threshold_static = cor_threshold_static,
                                                      cor_threshold_high = cor_threshold_high,
                                                      topLevel = FALSE,
                                                      verbose = verbose)
                                }
                                if (algorithm == "WKNN") {
                                    predRes <- WKNNcor(corMat = corMat,
                                                       subLevelModel = levelModel[[i]][[j]],
                                                       cutoff_method = cutoff_method,
                                                       k = k,
                                                       prob_threshold = prob_threshold,
                                                       cor_threshold_static = cor_threshold_static,
                                                       cor_threshold_high = cor_threshold_high,
                                                       topLevel = FALSE,
                                                       verbose = verbose)
                                }

                                if (algorithm == "DWKNN") {
                                    predRes <- DWKNNcor(corMat = corMat,
                                                        subLevelModel = levelModel[[i]][[j]],
                                                        cutoff_method = cutoff_method,
                                                        k = k,
                                                        prob_threshold = prob_threshold,
                                                        cor_threshold_static = cor_threshold_static,
                                                        cor_threshold_high = cor_threshold_high,
                                                        topLevel = FALSE,
                                                        verbose = verbose)
                                }

                                pred_level[[j]] <- predRes$predRes
                            }

                        }

                    }

                    # Else, the model of (level i-1, cells labeled as j) IS "no model"
                    # maintain the same class.
                    else {
                        predIdx <- which(pred[[i - 1]] == j)
                        # check the label of current level based on the label of last level
                        pred_level[[j]] <- as.factor(rep(unique(cutree_list[[i]][cutree_list[[i - 1]] == j]),
                                                         length(predIdx)))
                        names(pred_level[[j]]) <- colnames(exprsMat_test)[predIdx]
                    }
                }



                # Get the predict results for level i, change it to the numeric

                pred[[i]] <- unlist(lapply(pred_level, function(x){
                    cellNames <- names(x)
                    x <- as.numeric(as.character(x))
                    names(x) <- cellNames
                    x
                }))

                # reorder the prediction results to consistent
                #with the exprsMat_test
                pred[[i]] <- pred[[i]][colnames(exprsMat_test)]

                names(pred[[i]]) <- colnames(exprsMat_test)
                pred[[i]] <- as.numeric(as.character(pred[[i]]))
                # there will be NA since they are unassigned
                #from the last level,
                #and therefore are not predicted in this level
                pred[[i]][is.na(pred[[i]])] <- 0
                names(pred[[i]]) <- colnames(exprsMat_test)


            }
            else{

                # else, if all the cells in paraent nodes are unassigned
                pred[[i]] <- as.factor(rep(0, ncol(exprsMat_test)))
                names(pred[[i]]) <- colnames(exprsMat_test)
            }

        }else{
            # If this level is NULL (Level 1)
            pred[[i]] <- as.factor(rep(1, ncol(exprsMat_test)))
            names(pred[[i]]) <- colnames(exprsMat_test)
        }


    }

    predMat <- do.call(cbind, pred)
    predMat <- vapply(seq_len(ncol(predMat)),
                      function(x) getPredRes(predMat, cutree_list, x),
                      character(nrow(predMat)))

    predRes <- apply(predMat, 1, function(x){
        unAssIdx <- which(x == "unassigned")
        if (length(unAssIdx) != 0) {
            if (min(unAssIdx) >= 3) {
                x[min(unAssIdx) - 1]
            }else{
                "unassigned"
            }
        }else{
            x[length(x)]
        }
    })

    if (is.null(cellTypes_test)) {
        return(list(predRes = predRes, predLabelMat = predMat))
    }else{

        classify_res <-  ClassifyError(cellTypes_pred = predRes,
                                       cellTypes_test = cellTypes_test,
                                       cellTypes_train = cellTypes_train)

        if (verbose) {
            print(table(classify_res)/length(classify_res))
        }

        return(list(predRes = predRes, predLabelMat = predMat,
                    classifyRes = classify_res))

    }

}



# Function to calculate similarity
#' @importFrom proxyC simil
#' @importFrom proxy as.dist
#' @importFrom proxy dist
#' @importFrom methods as


calculateSimilarity <- function(exprsMat_train,
                                exprsMat_test,
                                similarity = c("pearson",  "spearman",
                                               "cosine", "jaccard", "kendall",
                                               "weighted_rank","manhattan")) {


    similarity <- match.arg(similarity, c("pearson",  "spearman",
                                          "cosine", "jaccard", "kendall",
                                          "weighted_rank","manhattan"))

    if ("dgCMatrix"  %in% is(exprsMat_test) &
        !"dgCMatrix" %in% is(exprsMat_train)) {
        exprsMat_train <- methods::as(exprsMat_train, "dgCMatrix")
    }

    if (!"dgCMatrix" %in% is(exprsMat_test) &
        "dgCMatrix" %in% is(exprsMat_train)) {
        exprsMat_test <- methods::as(exprsMat_test, "dgCMatrix")
    }


    if (similarity == "cosine") {

        if ("dgCMatrix" %in% is(exprsMat_test) &
            "dgCMatrix" %in% is(exprsMat_train)) {
            corMat <- proxyC::simil(Matrix::t(exprsMat_train),
                                    Matrix::t(exprsMat_test),
                                    method = "cosine")
        } else {
            corMat <- 1 - as.matrix(proxy::dist(t(as.matrix(exprsMat_train)),
                                                t(as.matrix(exprsMat_test)),
                                                method = "cosine"))
        }

        corMat[is.na(corMat) | is.infinite(corMat)] <- min(corMat)

    } else if (similarity == "kendall") {



        corMat <- stats::cor(as.matrix(exprsMat_train),
                             as.matrix(exprsMat_test), method = "kendall")
        corMat[is.na(corMat) | is.infinite(corMat)] <- -1

    } else if (similarity == "jaccard") {

        if ("dgCMatrix" %in% is(exprsMat_test) &
            "dgCMatrix" %in% is(exprsMat_train)) {
            corMat <- proxyC::simil(Matrix::t(exprsMat_train),
                                    Matrix::t(exprsMat_test),
                                    method = "jaccard")
        } else {
            corMat <- 1 -
                as.matrix(proxy::dist(t(as.matrix(exprsMat_train > 0)),
                                      t(as.matrix(exprsMat_test > 0)),
                                      method = "Jaccard"))

        }
        corMat[is.na(corMat) | is.infinite(corMat)] <- min(corMat)
    }else if (similarity == "weighted_rank") {

        corMat <- wtd_rank2(as.matrix(exprsMat_train),
                            as.matrix(exprsMat_test), method = "pearson")
        corMat[is.na(corMat) | is.infinite(corMat)] <- -1

    }else if (similarity == "manhattan") {
        if ("dgCMatrix" %in% is(exprsMat_test) &
            "dgCMatrix" %in% is(exprsMat_train)) {
            corMat <- 1 - as.matrix(proxy::dist(t(as.matrix(exprsMat_train)),
                                                t(as.matrix(exprsMat_test)),
                                                method = "Manhattan"))
        } else {
            corMat <- 1 - as.matrix(proxy::dist(t(as.matrix(exprsMat_train)),
                                                t(as.matrix(exprsMat_test)),
                                                method = "Manhattan"))
        }
        corMat[is.na(corMat) | is.infinite(corMat)] <- min(corMat)

    }else if (similarity == "spearman") {

        corMat <- stats::cor(as.matrix(exprsMat_train),
                             as.matrix(exprsMat_test), method = "spearman")
        corMat[is.na(corMat) | is.infinite(corMat)] <- -1

    }else if (similarity == "pearson") {

        if ("dgCMatrix" %in% is(exprsMat_test) &
            "dgCMatrix" %in% is(exprsMat_train)) {
            corMat <- proxyC::simil(Matrix::t(exprsMat_train),
                                    Matrix::t(exprsMat_test),
                                    method = "correlation")
        } else {
            corMat <- stats::cor(as.matrix(exprsMat_train),
                                 as.matrix(exprsMat_test), method = "pearson")
        }
        corMat[is.na(corMat) | is.infinite(corMat)] <- -1

    }else{

        if ("dgCMatrix" %in% is(exprsMat_test) &
            "dgCMatrix" %in% is(exprsMat_train)) {
            corMat <- proxyC::simil(Matrix::t(exprsMat_train),
                                    Matrix::t(exprsMat_test),
                                    method = "correlation")
        } else {
            corMat <- stats::cor(as.matrix(exprsMat_train),
                                 as.matrix(exprsMat_test), method = "pearson")
        }
        corMat[is.na(corMat) | is.infinite(corMat)] <- -1
    }

    return(corMat)
}


# Function to caculate weighted rank correlation
# the codes are modified from the wtd_rank function in dismay package

wtd_rank2 <- function(mat1, mat2 = NULL, method = "pearson") {

    method <- match.arg(method, c("pearson", "spearman"))

    ranks1 <- apply(mat1, 2, function(x) rank(-x, ties.method = "average"))
    # weight the ranks
    # calculate the savage scores
    n1 <- nrow(mat1)
    reciprocals1 <- 1 / seq_len(n1)
    savage1 <- vapply(seq_len(n1),
                      function(i) sum(reciprocals1[i:n1]),
                      numeric(1L))
    # replace each rank with the savage score
    savages1 <- ranks1
    savages1[] <- savage1[ranks1]

    if (!is.null(mat2)) {
        ranks2 <- apply(mat2, 2, function(x) rank(-x, ties.method = "average"))
        # weight the ranks
        # calculate the savage scores
        n2 <- nrow(mat2)
        reciprocals2 <- 1 / seq_len(n2)
        savage2 <- vapply(seq_len(n2),
                          function(i) sum(reciprocals2[i:n2]),
                          numeric(1L))
        # replace each rank with the savage score
        savages2 <- ranks2
        savages2[] <- savage2[ranks2]

        cor <- stats::cor(savages1, savages2, method = method)
    } else {
        cor <- stats::cor(savages1, method = method)
    }


    # calculate pearson correlation

    return(cor)
}

# Function to get the prediction labels according to the tree
getPredRes <- function(predMat, cutree_list, level){
    res <- predMat[,level]
    for (i in seq_len(length(predMat[,level]))) {
        if (predMat[i,level] == 0) {
            res[i] <- "unassigned"
        }else{
            res[i] <- paste(names(cutree_list[[level]])[cutree_list[[level]]
                                                        %in% predMat[i,level]],
                            collapse = "_")
        }

    }
    return(res)
}


# Functions to classify the error for the predicted cell types

ClassifyError <- function(cellTypes_pred, cellTypes_test, cellTypes_train){

    errClass <- c("correct", "correctly unassigned",
                  "intermediate", "incorrectly unassigned",
                  "error assigned", "misclassified")


    if (length(cellTypes_pred) != length(cellTypes_test)) {
        stop("wrong input")
    }
    train_ref <- unique(cellTypes_train)
    res <- vapply(seq_len(length(cellTypes_pred)), function(i){
        if (cellTypes_test[i] %in% train_ref) {
            if (cellTypes_pred[i] %in% c("unassigned", "Unassigned")) {
                "incorrectly unassigned"
            } else if (cellTypes_pred[i] == "intermediate") {
                "intermediate"
            }else{
                if (cellTypes_test[i] == cellTypes_pred[i]) {
                    "correct"
                }else if (grepl(cellTypes_test[i], cellTypes_pred[i])) {
                    "intermediate"
                }
                else{
                    "misclassified"
                }
            }
        }else{
            if (cellTypes_pred[i] %in% c("unassigned","Unassigned")) {
                "correctly unassigned"
            }else{
                "error assigned"
            }
        }
    }, character(1L))

    res <- factor(res, levels = errClass)
    return(res)
}

setClassUnion("numeric_OR_NULL", members = c("numeric","NULL"))


#' An S4 class to stored training model for scClassify
#'
#' @slot name Name of the training dataset
#' @slot cellTypeTrain A vector of cell type in training dataset
#' @slot cellTypeTree A list indicate a cell type tree
#' @slot features A vector of character indicates the
#' features that are trained for this data
#' @slot model A list stored the training model,
#' including the features that are selected
#' and the cell expression matrix that are used for training
#' @slot modelweights A vector of numeric indicates the weights of each model
#' @slot metaData A DataFrame stored meta data of training model
#'
#' @importClassesFrom S4Vectors character_OR_NULL
#' @importClassesFrom S4Vectors DataFrame
#'
#'
#' @export
setClass("scClassifyTrainModel",
         slots = c(
             name = "character_OR_NULL",
             cellTypeTrain = "character",
             cellTypeTree = "list",
             features = "character",
             model = "list",
             modelweights = "numeric_OR_NULL",
             metaData = "DataFrame"))



#' The scClassifyTrainModel class
#'
#' The scClassifyTrainModel class is designed to stored
#' training model for scClassify
#' @param name Name of the training dataset
#' @param cellTypeTrain A vector of cell type in training dataset
#' @param cellTypeTree A list indicate a cell type tree
#' @param features A vector of character indicates the
#' features that are trained for this data
#' @param model A list stored the training model,
#' including the features that are selected
#' and the cell expression matrix that are used for training
#' @param modelweights A vector of numeric indicates the weights of each model
#' @param metaData A DataFrame stored meta data of training model
#'
#' @return A scClassifyTrainModel object
#'
#' @author Yingxin Lin
#'
#'
#' @importFrom methods new

.scClassifyTrainModel <- function(name,
                                  cellTypeTree,
                                  cellTypeTrain,
                                  features,
                                  model,
                                  modelweights,
                                  metaData) {

    scClassify_train_obj <- new("scClassifyTrainModel",
                                name = name,
                                cellTypeTree = cellTypeTree,
                                cellTypeTrain = cellTypeTrain,
                                features = features,
                                model = model,
                                modelweights = modelweights,
                                metaData = metaData)
    names(scClassify_train_obj@modelweights) <- names(modelweights)
    return(scClassify_train_obj)

}




setMethod("show", "scClassifyTrainModel",
          function(object) {
              cat(paste("Class:", class(object), "\n"))
              cat(paste("Model name:", object@name), "\n")
              cat("Feature selection methods: ")
              cat(object@features, "\n")
              cat("Number of cells in the training data: ")
              cat(length(object@cellTypeTrain), "\n")
              cat("Number of cell types in the training data: ")
              cat(length(object@cellTypeTree[[length(object@cellTypeTree)]]),
                  "\n")
          })



#' An S4 class to stored a list of training models from scClassify
#'
#'
#' @importClassesFrom S4Vectors SimpleList
#' @export

setClass("scClassifyTrainModelList",
         contains = "SimpleList")



#' The scClassifyTrainModelList class
#'
#' @param ... scClassifyTrainModel objects
#'
#' @return A scClassifyTrainModelList object
#'
#' @examples
#'
#' data("trainClassExample_xin")
#' data("trainClassExample_wang")
#' trainClassExampleList <- scClassifyTrainModelList(trainClassExample_xin,
#' trainClassExample_wang
#' )
#'
#' @importFrom S4Vectors SimpleList
#'
#' @export
scClassifyTrainModelList <- function(...) {
    l <- new("scClassifyTrainModelList", S4Vectors::SimpleList(...))
    name_list <- lapply(l@listData, function(x) x@name)
    if (any(duplicated(name_list))) {
        dup_names <- unique(name_list[duplicated(name_list)])
        for (i in seq_along(dup_names)) {
            idx_to_change <- which(name_list %in% dup_names[i])
            name_list[idx_to_change] <- paste(name_list[idx_to_change],
                                              seq_along(idx_to_change),
                                              sep = "_")
        }
    }
    names(l@listData) <- name_list
    l
}




setMethod("show", "scClassifyTrainModelList", function(object) {
    cat(paste("Class:", class(object), "\n"))
    cat("Number of Trainin model: ")
    cat(length(object), "\n")

    cat("Training data name:",
        names(object@listData), "\n")
})


#' Accessors of cellTypeTree for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage cellTypeTree(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' cellTypeTree(trainClassExample_xin)
#'
#' @return cellTypeTree of the scClassifyTrainModel slot
#'
#' @aliases
#' cellTypeTree,scClassifyTrainModel-method
#' cellTypeTree
#'
#' @export
setGeneric("cellTypeTree", function(x)
    standardGeneric("cellTypeTree"))
setMethod("cellTypeTree", "scClassifyTrainModel",
          function(x) {
              x@cellTypeTree
          })




setGeneric("cellTypeTree<-", function(x, value)
    standardGeneric("cellTypeTree<-"))
setReplaceMethod("cellTypeTree", "scClassifyTrainModel",
                 function(x, value) {
                     x@cellTypeTree <- value
                     x
                 })



#' Accessors of features for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage features(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' features(trainClassExample_xin)
#'
#' @return features of the scClassifyTrainModel slot
#'
#' @aliases
#' features,scClassifyTrainModel-method
#' features
#'
#' @export
setGeneric("features", function(x) standardGeneric("features"))
setMethod("features", "scClassifyTrainModel", function(x) {
    x@features
})



setGeneric("features<-", function(x, value)
    standardGeneric("features<-"))
setReplaceMethod("features", "scClassifyTrainModel",
                 function(x, value) {
                     x@features <- value
                     x
                 })








#' Accessors of name for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage name(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' name(trainClassExample_xin)
#'
#' @return name of the scClassifyTrainModel slot
#'
#' @aliases
#' name,scClassifyTrainModel-method
#' name
#'
#'
#' @export
setGeneric("name", function(x) standardGeneric("name"))
setMethod("name", "scClassifyTrainModel", function(x) {
    x@name
})



setGeneric("name<-", function(x, value)
    standardGeneric("name<-"))
setReplaceMethod("name", "scClassifyTrainModel",
                 function(x, value) {
                     x@name <- value
                     x
                 })








#' Accessors of cellTypeTrain for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage cellTypeTrain(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' cellTypeTrain(trainClassExample_xin)
#'
#' @return cellTypeTrain of the scClassifyTrainModel slot
#'
#' @aliases
#' cellTypeTrain,scClassifyTrainModel-method
#' cellTypeTrain
#'
#'
#' @export
setGeneric("cellTypeTrain", function(x) standardGeneric("cellTypeTrain"))
setMethod("cellTypeTrain", "scClassifyTrainModel", function(x) {
    x@cellTypeTrain
})



setGeneric("cellTypeTrain<-", function(x, value)
    standardGeneric("cellTypeTrain<-"))
setReplaceMethod("cellTypeTrain", "scClassifyTrainModel",
                 function(x, value) {
                     x@cellTypeTrain <- value
                     x
                 })








#' Accessors of modelweights for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage modelweights(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' modelweights(trainClassExample_xin)
#'
#' @return modelweights of the scClassifyTrainModel slot
#'
#'
#' @aliases
#' modelweights,scClassifyTrainModel-method
#' modelweights
#'
#'
#' @export
setGeneric("modelweights", function(x) standardGeneric("modelweights"))
setMethod("modelweights", "scClassifyTrainModel", function(x) {
    x@modelweights
})



setGeneric("modelweights<-", function(x, value)
    standardGeneric("modelweights<-"))
setReplaceMethod("modelweights", "scClassifyTrainModel",
                 function(x, value) {
                     x@modelweights <- value
                     x
                 })






#' Accessors of model for scClassifyTrainModel
#'
#' Methods to access various components of the `scClassifyTrainModel` object.
#'
#' @usage model(x)
#'
#' @param x A `scClassifyTrainModel` object.
#'
#' @return model of the scClassifyTrainModel slot
#'
#' @examples
#'
#' data(trainClassExample_xin)
#' model(trainClassExample_xin)
#'
#' @aliases
#' model,scClassifyTrainModel-method
#' model
#'
#'
#' @export
setGeneric("model", function(x) standardGeneric("model"))
setMethod("model", "scClassifyTrainModel", function(x) {
    x@model
})



setGeneric("model<-", function(x, value)
    standardGeneric("model<-"))
setReplaceMethod("model", "scClassifyTrainModel",
                 function(x, value) {
                     x@model <- value
                     x
                 })

#' @title Train and test scClassify model
#'
#' @param exprsMat_train A matrix of log-transformed expression matrix of reference dataset
#' @param cellTypes_train A vector of cell types of reference dataset
#' @param exprsMat_test A list or a matrix indicates the expression matrices of the query datasets
#' @param cellTypes_test A list or a vector indicates cell types of the query datasets (Optional).
#' @param tree A vector indicates the method to build hierarchical tree, set as "HOPACH" by default.
#' This should be one of "HOPACH" and "HC" (using hclust).
#' @param selectFeatures A vector indicates the gene selection method, set as "limma" by default.
#' This should be one or more of "limma", "DV", "DD", "chisq", "BI" and "Cepo".
#' @param algorithm A vector indicates the KNN method that are used, set as
#' "WKNN" by default. Thisshould be one or more of "WKNN", "KNN", "DWKNN".
#' @param similarity A vector indicates the similarity measure that are used,
#' set as "pearson" by default. This should be one or more of "pearson",
#' "spearman", "cosine", "jaccard", kendall", "binomial", "weighted_rank","manhattan"
#' @param cutoff_method A vector indicates the method to cutoff the correlation distribution.
#' Set as "dynamic" by default.
#' @param weighted_ensemble A logical input indicates in ensemble learning,
#' whether the results is combined by a weighted score for each base classifier.
#' @param weights A vector indicates the weights for ensemble
#' @param weighted_jointClassification A logical input indicates in joint classification
#' using multiple training datasets,
#' whether the results is combined by a weighted score for each training model.
#' @param cellType_tree A list indicates the cell type tree provided by user.
#' (By default, it is NULL) (Only for one training data input)
#' @param k An integer indicates the number of neighbour
#' @param topN An integer indicates the top number of features that are selected
#' @param hopach_kmax An integer between 1 and 9 specifying the maximum number of
#' children at each node in the HOPACH tree.
#' @param pSig A numeric indicates the cutoff of pvalue for features
#' @param prob_threshold A numeric indicates the probability threshold for KNN/WKNN/DWKNN.
#' @param cor_threshold_static A numeric indicates the static correlation threshold.
#' @param cor_threshold_high A numeric indicates the highest correlation threshold
#' @param returnList A logical input indicates whether the output will be class of list
#' @param parallel A logical input indicates whether running in paralllel or not
#' @param BPPARAM  A \code{BiocParallelParam} class object
#' from the \code{BiocParallel} package is used. Default is SerialParam().
#' @param verbose A logical input indicates whether the intermediate steps will be printed
#'
#' @return A list of the results, including testRes storing the results of the testing information,
#' and trainRes storing the training model inforamtion.
#'
#' @author Yingxin Lin
#'
#' @examples
#'
#' data("scClassify_example")
#' xin_cellTypes <- scClassify_example$xin_cellTypes
#' exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
#' wang_cellTypes <- scClassify_example$wang_cellTypes
#' exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
#'
#' scClassify_res <- scClassify(exprsMat_train = exprsMat_xin_subset,
#' cellTypes_train = xin_cellTypes,
#' exprsMat_test = list(wang = exprsMat_wang_subset),
#' cellTypes_test = list(wang = wang_cellTypes),
#' tree = "HOPACH",
#' algorithm = "WKNN",
#' selectFeatures = c("limma"),
#' similarity = c("pearson"),
#' returnList = FALSE,
#' verbose = FALSE)
#'
#' @importFrom S4Vectors DataFrame
#' @importFrom methods is
#' @importFrom BiocParallel SerialParam
#' @export


scClassify <- function(exprsMat_train = NULL,
                       cellTypes_train = NULL,
                       exprsMat_test = NULL,
                       cellTypes_test = NULL,
                       tree = "HOPACH",
                       algorithm = "WKNN",
                       selectFeatures = "limma",
                       similarity = "pearson",
                       cutoff_method = c("dynamic", "static"),
                       weighted_ensemble = FALSE,
                       weights = NULL,
                       weighted_jointClassification = TRUE,
                       cellType_tree = NULL,
                       k = 10,
                       topN = 50,
                       hopach_kmax = 5,
                       pSig = 0.01,
                       prob_threshold = 0.7,
                       cor_threshold_static = 0.5,
                       cor_threshold_high = 0.7,
                       returnList = TRUE,
                       parallel = FALSE,
                       BPPARAM = BiocParallel::SerialParam(),
                       verbose = FALSE) {


    # check input
    if (is.null(exprsMat_train) | is.null(cellTypes_train) |
        is.null(exprsMat_test)) {
        stop("exprsMat_train or cellTypes_train or exprsMat_test is NULL!")
    }

    if (!is.null(cellTypes_test)) {
        if ("character" %in% is(cellTypes_test)) {
            if (length(cellTypes_test) != ncol(exprsMat_test)) {
                stop("Length of testing cell types does not match
             with number of column of testing expression matrix")
            }
        }

        if ("list" %in% is(cellTypes_test)) {
            if (sum(unlist(lapply(cellTypes_test, length)) !=
                    unlist(lapply(exprsMat_test, ncol))) != 0) {
                stop("Length of testing cell types does not match
             with number of column of testing expression matrix")
            }
        }

    }

    if ("list" %in% is(exprsMat_train) ) {
        if (sum(unlist(lapply(cellTypes_train, length)) !=
                unlist(lapply(exprsMat_train, ncol))) != 0) {
            stop("Length of training cell types does not match with
           number of column of training expression matrix")
        }
    }else {
        if (length(cellTypes_train) != ncol(exprsMat_train)) {
            stop("Length of training cell types does not match with
           number of column of training expression matrix")
        }
    }

    if ("list" %in% is(exprsMat_train)) {
        if (any(lapply(cellTypes_train, function(x) any(table(x) == 1)))) {
            stop("There is cell type with only one cell,
           please check cellTypes_train")
        }
    }else {
        if (any(table(cellTypes_train) == 1)) {
            stop("There is cell type with only one cell,
           please check cellTypes_train")
        }
    }




    tree <- match.arg(tree, c("HOPACH", "HC"), several.ok = FALSE)
    selectFeatures <- match.arg(selectFeatures,
                                c("limma", "DV", "DD", "chisq", "BI", "Cepo"),
                                several.ok = TRUE)

    algorithm <- match.arg(algorithm,
                           c("WKNN", "KNN", "DWKNN"),
                           several.ok = TRUE)

    similarity <- match.arg(similarity, c("pearson",  "spearman",
                                          "cosine", "jaccard", "kendall",
                                          "weighted_rank","manhattan"),
                            several.ok = TRUE)
    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"))



    # To check if need to run weighted ensemble learning

    if ((length(selectFeatures) > 1 | length(algorithm) > 1 |
         length(similarity) > 1) ) {
        if (weighted_ensemble) {
            weighted_ensemble <- TRUE
            ensemble <- TRUE

            if (verbose) {
                cat("Performing weighted ensemble learning... \n")
            }

        } else {
            weighted_ensemble <- FALSE
            ensemble <- TRUE

            if (verbose) {
                cat("Performing unweighted ensemble learning... \n")
            }

        }
    } else {
        weighted_ensemble <- FALSE
        ensemble <- FALSE

        if (verbose) {
            cat("Ensemble learning is disabled... \n")
        }

    }

    # To check if need to run weighted joint classification

    if ("list" %in% is(exprsMat_train) &
        length(exprsMat_train) > 1 &
        weighted_jointClassification) {
        cat("Performing weighted joint classification \n")
        weighted_jointClassification <- TRUE
    } else {
        weighted_jointClassification <- FALSE
    }



    ensemble_methods <- as.matrix(expand.grid(similarity = similarity,
                                              algorithm = algorithm,
                                              features = selectFeatures))

    if (!is.null(weights)) {
        if (length(weights) != nrow(ensemble_methods)) {
            stop("The length of weights is not equal to
           the number of combination of ensemble methods")
        }
    }



    # calculate the weights for train model
    if (weighted_jointClassification | (weighted_ensemble & is.null(weights))) {
        weightsCal = TRUE
    } else {
        weightsCal = FALSE
    }



    ### train_scClassify
    trainRes <- train_scClassify(exprsMat_train,
                                 cellTypes_train,
                                 tree = tree,
                                 selectFeatures = selectFeatures,
                                 topN = topN,
                                 hopach_kmax = hopach_kmax,
                                 pSig = pSig,
                                 cellType_tree = cellType_tree,
                                 weightsCal = weightsCal,
                                 parallel = parallel,
                                 BPPARAM = BPPARAM,
                                 verbose = verbose,
                                 k = k,
                                 prob_threshold = prob_threshold,
                                 cor_threshold_static = cor_threshold_static,
                                 cor_threshold_high = cor_threshold_high,
                                 algorithm = algorithm,
                                 similarity = similarity,
                                 cutoff_method = cutoff_method
    )





    if (verbose) {
        cat("Predicting using followings parameter combinations: \n")
        print(ensemble_methods)
    }

    ### if there are multiple testing datasets
    if (verbose) {
        cat("==================  Start classifying on test dataset  ==== \n")
    }

    if ("list" %in% is(exprsMat_test)) {
        testRes <- list()

        for (testDataset_idx in seq_len(length(exprsMat_test))) {

            if (verbose) {
                cat("Predicting: ")
                print(names(exprsMat_test)[testDataset_idx])
            }

            if ("list" %in% is(exprsMat_train)) {
                # for the case there are multiple training datasets
                #
                predictRes <- list()


                for (train_list_idx in seq_len(length(exprsMat_train))) {

                    if (verbose) {
                        cat("Training using ")
                        cat(names(exprsMat_train)[train_list_idx], "\n")
                    }

                    predictRes[[train_list_idx]] <- predict_scClassify(exprsMat_test = exprsMat_test[[testDataset_idx]],
                                                                       trainRes  = trainRes[[train_list_idx]],
                                                                       cellTypes_test = cellTypes_test[[testDataset_idx]],
                                                                       k = k,
                                                                       prob_threshold = prob_threshold,
                                                                       cor_threshold_static = cor_threshold_static,
                                                                       cor_threshold_high = cor_threshold_high,
                                                                       algorithm = algorithm,
                                                                       features = selectFeatures,
                                                                       similarity = similarity,
                                                                       cutoff_method = cutoff_method,
                                                                       weighted_ensemble = weighted_ensemble,
                                                                       weights = weights,
                                                                       parallel = parallel,
                                                                       BPPARAM = BPPARAM,
                                                                       verbose = verbose)

                }
                names(predictRes) <- paste("Trained_by",
                                           names(trainRes), sep = "_")


                jointRes <- getJointRes(predictRes, trainRes)

                rownames(jointRes) <- colnames(exprsMat_test)


                predictRes$jointRes <- jointRes


            }else {
                predictRes <- predict_scClassify(exprsMat_test = exprsMat_test[[testDataset_idx]],
                                                 trainRes  = trainRes,
                                                 cellTypes_test = cellTypes_test[[testDataset_idx]],
                                                 k = k,
                                                 prob_threshold = prob_threshold,
                                                 cor_threshold_static = cor_threshold_static,
                                                 cor_threshold_high = cor_threshold_high,
                                                 algorithm = algorithm,
                                                 features = selectFeatures,
                                                 similarity = similarity,
                                                 cutoff_method = cutoff_method,
                                                 weighted_ensemble = weighted_ensemble,
                                                 weights = weights,
                                                 parallel = parallel,
                                                 BPPARAM = BPPARAM,
                                                 verbose = verbose)


            }

            testRes[[testDataset_idx]] <- predictRes
        }

        names(testRes) <- names(exprsMat_test)
    }else{
        # else only one dataset as a matrix in the test
        if ("list" %in% is(exprsMat_train)) {
            # for the case there are multiple training datasets
            #
            testRes <- list()
            for (train_list_idx in seq_len(length(exprsMat_train))) {

                if (verbose) {
                    cat("Training using ")
                    cat(names(exprsMat_train)[train_list_idx], "\n")
                }
                testRes[[train_list_idx]] <- predict_scClassify(exprsMat_test = exprsMat_test,
                                                                trainRes  = trainRes[[train_list_idx]],
                                                                cellTypes_test = cellTypes_test,
                                                                k = k,
                                                                prob_threshold = prob_threshold,
                                                                cor_threshold_static = cor_threshold_static,
                                                                cor_threshold_high = cor_threshold_high,
                                                                algorithm = algorithm,
                                                                features = selectFeatures,
                                                                similarity = similarity,
                                                                cutoff_method = cutoff_method,
                                                                weighted_ensemble = weighted_ensemble,
                                                                weights = weights,
                                                                parallel = parallel,
                                                                BPPARAM = BPPARAM,
                                                                verbose = verbose)

            }


            names(testRes) <- paste("Trained_by", names(trainRes), sep = "_")

            jointRes <- getJointRes(testRes, trainRes)

            rownames(jointRes) <- colnames(exprsMat_test)


            testRes$jointRes <- jointRes

        }else {
            predictRes <- predict_scClassify(exprsMat_test = exprsMat_test,
                                             trainRes  = trainRes,
                                             cellTypes_test = cellTypes_test,
                                             k = k,
                                             prob_threshold = prob_threshold,
                                             cor_threshold_static = cor_threshold_static,
                                             cor_threshold_high = cor_threshold_high,
                                             algorithm = algorithm,
                                             features = selectFeatures,
                                             similarity = similarity,
                                             cutoff_method = cutoff_method,
                                             weighted_ensemble = weighted_ensemble,
                                             weights = weights,
                                             parallel = parallel,
                                             BPPARAM = BPPARAM,
                                             verbose = verbose)
            testRes <- list(test = predictRes)
        }




    }


    if (returnList) {
        return(list(testRes = testRes, trainRes = trainRes))
    } else {

        if ("list" %in% is(exprsMat_train)) {
            trainClassList <- list()
            for (train_list_idx in seq_len(length(trainRes))) {
                trainClassList[[train_list_idx]] <- .scClassifyTrainModel(
                    name = names(trainRes)[train_list_idx],
                    cellTypeTree = trainRes[[train_list_idx]]$cutree_list,
                    cellTypeTrain = as.character(trainRes[[train_list_idx]]$cellTypes_train),
                    features = names(trainRes[[train_list_idx]]$hierarchyKNNRes),
                    model = trainRes[[train_list_idx]]$hierarchyKNNRes,
                    modelweights = trainRes[[train_list_idx]]$modelweights,
                    metaData = S4Vectors::DataFrame())

            }
            trainClassList <- scClassifyTrainModelList(trainClassList)
        } else {
            trainClassList <- .scClassifyTrainModel(
                name = "training",
                cellTypeTree = trainRes$cutree_list,
                cellTypeTrain = as.character(trainRes$cellTypes_train),
                features = names(trainRes$hierarchyKNNRes),
                model = trainRes$hierarchyKNNRes,
                modelweights = trainRes$modelweights,
                metaData = S4Vectors::DataFrame())
        }


        return(list(testRes = testRes, trainRes = trainClassList))


    }


}

#' @title Example data used in scClassify package
#' @description  A list includes expression matrix and cell type of subsets of
#' wang et al., xin et al.
#'
#' @usage data(scClassify_example, package = 'scClassify')
#' @source
#' Wang YJ, Schug J, Won K-J, Liu C, Naji A, Avrahami D, Golson ML &
#' Kaestner KH (2016) Single cell transcriptomics of the human endocrine
#' pancreas. Diabetes: db160405
#'
#' Xin Y, Kim J, Okamoto H, Ni M, Wei Y, Adler C, Murphy AJ, Yancopoulos GD,
#' Lin C & Gromada J (2016) RNA sequencing of single human islet cells reveals
#' type 2 diabetes genes. Cell Metab. 24: 608–615
"scClassify_example"


#' @title Subset of pretrained model of Xin et al.
#' @description  An obejct of scClassifyTrainModel for Xin et al.
#' @usage data(trainClassExample_xin, package = 'scClassify')
#' @source Xin Y, Kim J, Okamoto H, Ni M, Wei Y, Adler C, Murphy AJ, Yancopoulos GD,
#' Lin C & Gromada J (2016) RNA sequencing of single human islet cells reveals
#' type 2 diabetes genes. Cell Metab. 24: 608–615
"trainClassExample_xin"

#' @title Subset of pretrained model of Wang et al.
#' @description  An obejct of scClassifyTrainModel for Wang et al.
#' @usage data(trainClassExample_wang, package = 'scClassify')
#'
#' @source Wang YJ, Schug J, Won K-J, Liu C, Naji A, Avrahami D, Golson ML &
#' Kaestner KH (2016) Single cell transcriptomics of the human endocrine
#' pancreas. Diabetes: db160405
#'
"trainClassExample_wang"

#' Training scClassify model
#'
#' @param exprsMat_train A matrix of log-transformed expression matrix of reference dataset
#' @param cellTypes_train A vector of cell types of reference dataset
#' @param tree A vector indicates the method to build hierarchical tree,
#' set as "HOPACH" by default.
#' This should be one of "HOPACH" and "HC" (using stats::hclust).
#' @param selectFeatures A vector indicates the gene selection method,
#' set as "limma" by default.
#' This should be one or more of "limma", "DV", "DD", "chisq", "BI", "Cepo".
#' @param topN An integer indicates the top number of features that are selected
#' @param hopach_kmax An integer between 1 and 9 specifying the maximum number of
#' children at each node in the HOPACH tree.
#' @param pSig A numeric indicates the cutoff of pvalue for features
#' @param cellType_tree A list indicates the cell type tree provided by user.
#' (By default, it is NULL)
#' @param weightsCal A logical input indicates whether we need to
#' calculate the weights for the model.
#' @param parallel A logical input indicates whether the algorihms will run in parallel
#' @param BPPARAM  A \code{BiocParallelParam} class object
#' from the \code{BiocParallel} package is used. Default is SerialParam().
#' @param verbose A logical input indicates whether the intermediate steps will be printed
#' @param returnList A logical input indicates whether the output will be class of list
#' @param ... Other input for predict_scClassify for the case when weights calculation
#' of the pretrained model is performed
#' @return list of results or an object of \code{scClassifyTrainModel}
#' @author Yingxin Lin
#'
#' @examples
#' data("scClassify_example")
#' xin_cellTypes <- scClassify_example$xin_cellTypes
#' exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
#' trainClass <- train_scClassify(exprsMat_train = exprsMat_xin_subset,
#' cellTypes_train = xin_cellTypes,
#' selectFeatures = c("limma", "BI"),
#' returnList = FALSE
#' )
#'
#' @importFrom stats na.omit
#' @importFrom methods is
#' @importFrom BiocParallel SerialParam
#' @export




train_scClassify <- function(exprsMat_train,
                             cellTypes_train,
                             tree = "HOPACH",
                             selectFeatures = "limma",
                             topN = 50,
                             hopach_kmax = 5,
                             pSig = 0.05,
                             cellType_tree = NULL,
                             weightsCal = FALSE,
                             parallel= FALSE,
                             BPPARAM = BiocParallel::SerialParam(),
                             verbose= TRUE,
                             returnList = TRUE,
                             ...){


  if (is.null(exprsMat_train) | is.null(cellTypes_train)) {
    stop("exprsMat_train or cellTypes_train or exprsMat_test is NULL!")
  }

  # Matching the argument of the tree construction method
  tree <- match.arg(tree, c("HOPACH", "HC"), several.ok = FALSE)

  # Matching the argument of feature selection method
  selectFeatures <- match.arg(selectFeatures,
                              c("limma", "DV", "DD", "chisq", "BI", "Cepo"),
                              several.ok = TRUE)


  if ("list" %in% is(exprsMat_train)) {
    if (sum(unlist(lapply(cellTypes_train, length)) !=
            unlist(lapply(exprsMat_train, ncol))) != 0) {
      stop("Length of training cell types does not match with
           number of column of training expression matrix")
    }
  }else {
    if (length(cellTypes_train) != ncol(exprsMat_train)) {
      stop("Length of training cell types does not match with
           number of column of training expression matrix")
    }
  }




  # To rename the train list if name is null (only when there are multiple training datasets)
  if ( "list" %in% is(exprsMat_train)) {
    if (is.null(names(exprsMat_train))) {
      names(exprsMat_train) <- names(cellTypes_train) <-
        paste("TrainData",
              seq_len(length(exprsMat_train)),
              sep = "_")
    } else if (sum(names(exprsMat_train) == "") != 0) {
      names(exprsMat_train)[names(exprsMat_train) == ""] <-
        names(cellTypes_train)[names(cellTypes_train) == ""] <-
        paste("TrainData", which(names(exprsMat_train) == ""), sep = "_")
    }
  }

  # QC for the training data set

  if (any(c("matrix", "dgCMatrix") %in% is(exprsMat_train))) {

    zeros <- apply(exprsMat_train, 1, function(x) sum(x == 0)/length(x))
    minPctCell <- min(table(cellTypes_train)/length(cellTypes_train))
    exprsMat_train <- exprsMat_train[zeros <= max(1 - minPctCell, 0.95), ]
    if (verbose) {
      cat("after filtering not expressed genes \n")
      print(dim(exprsMat_train))
    }
  } else {
    for (train_list_idx in seq_len(length(exprsMat_train))) {
      zeros <- apply(exprsMat_train[[train_list_idx]], 1,
                     function(x) sum(x == 0)/length(x))
      minPctCell <- min(table(cellTypes_train[[train_list_idx]])/length(cellTypes_train[[train_list_idx]]))
      exprsMat_train[[train_list_idx]] <- exprsMat_train[[train_list_idx]][zeros <= max(1 - minPctCell, 0.95), ]
    }
    if (verbose) {
      cat("after filtering not expressed genes \n")
      print(lapply(exprsMat_train, dim))
    }
  }

  ### train_scClassify
  if ("list" %in% is(exprsMat_train)) {
    trainRes <- list()
    for (train_list_idx in seq_len(length(exprsMat_train))) {
      trainRes[[train_list_idx]] <- train_scClassifySingle(exprsMat_train[[train_list_idx]],
                                                           cellTypes_train[[train_list_idx]],
                                                           tree = tree,
                                                           selectFeatures = selectFeatures,
                                                           topN = topN,
                                                           hopach_kmax = hopach_kmax,
                                                           pSig = pSig,
                                                           weightsCal = weightsCal,
                                                           parallel = parallel,
                                                           BPPARAM = BPPARAM,
                                                           verbose = verbose,
                                                           ...)
    }
    names(trainRes) <- names(exprsMat_train)
  } else{

    trainRes <- train_scClassifySingle(exprsMat_train,
                                       cellTypes_train,
                                       tree = tree,
                                       selectFeatures = selectFeatures,
                                       topN = topN,
                                       hopach_kmax = hopach_kmax,
                                       pSig = pSig,
                                       cellType_tree = cellType_tree,
                                       weightsCal = weightsCal,
                                       parallel = parallel,
                                       BPPARAM = BPPARAM,
                                       verbose = verbose,
                                       ...)
  }





  # return the results
  if (returnList) {

    return(trainRes)

  } else {
    if ("list" %in% is(exprsMat_train)) {
      trainClassList <- list()
      for (train_list_idx in seq_len(length(trainRes))) {
        trainClassList[[train_list_idx]] <- .scClassifyTrainModel(
          name = names(trainRes)[train_list_idx],
          cellTypeTree = trainRes[[train_list_idx]]$cutree_list,
          cellTypeTrain = as.character(trainRes[[train_list_idx]]$cellTypes_train),
          features = names(trainRes[[train_list_idx]]$hierarchyKNNRes),
          model = trainRes[[train_list_idx]]$hierarchyKNNRes,
          modelweights = trainRes[[train_list_idx]]$modelweights,
          metaData = S4Vectors::DataFrame())

      }
      trainClassList <- scClassifyTrainModelList(trainClassList)
    } else {
      trainClassList <- .scClassifyTrainModel(
        name = "training",
        cellTypeTree = trainRes$cutree_list,
        cellTypeTrain = as.character(trainRes$cellTypes_train),
        features = names(trainRes$hierarchyKNNRes),
        model = trainRes$hierarchyKNNRes,
        modelweights = trainRes$modelweights,
        metaData = S4Vectors::DataFrame())
    }

    return(trainClassList)

  }


}




#' @importFrom BiocParallel SerialParam bplapply

train_scClassifySingle <- function(exprsMat_train,
                                   cellTypes_train,
                                   tree = "HOPACH",
                                   selectFeatures = "limma",
                                   topN = 50,
                                   hopach_kmax = 5,
                                   pSig = 0.05,
                                   cellType_tree = NULL,
                                   weightsCal = FALSE,
                                   parallel= FALSE,
                                   BPPARAM = BiocParallel::SerialParam(),
                                   verbose= TRUE,
                                   ...){

  if (is.null(rownames(exprsMat_train))) {
    stop("rownames of the exprsMat_train is NULL!")
  }

  if (is.null(rownames(exprsMat_train)) |
      sum(duplicated(colnames(exprsMat_train))) != 0) {
    stop("colnames of exprsMat_train is NULL or not unique")
  }

  if (length(cellTypes_train) != ncol(exprsMat_train)) {
    stop("Length of training cell types does not match with
         number of column of training expression matrix")
  }

  if (all(exprsMat_train %% 1 == 0)) {
    warning("exprsMat_train looks like a count matrix
            (scClassify requires a log-transformed normalised input)")
  }

  # Matching the argument of the tree construction method
  tree <- match.arg(tree, c("HOPACH", "HC"), several.ok = FALSE)

  # Matching the argument of feature selection method
  selectFeatures <- match.arg(selectFeatures,
                              c("limma", "DV", "DD", "chisq", "BI", "Cepo"),
                              several.ok = TRUE)


  if (verbose) {
    print("Feature Selection...")
  }


  # Select the features to construct tree
  tt <- doLimma(exprsMat_train, cellTypes_train)
  de <- Reduce(union, lapply(tt, function(t)
    rownames(t)[seq_len(max(min(50, sum(t$adj.P.Val < 0.001)), 30))]))
  de <- na.omit(de)
  if (verbose) {
    print(paste("Number of genes selected to construct HOPACH tree",
                length(de)))
  }


  if (is.null(cellType_tree)) {
    # Calculate the centroid matrix for tree
    # construction using selected features
    centroidMat <- do.call(cbind, lapply(unique(cellTypes_train), function(x)
      Matrix::rowMeans(as.matrix(exprsMat_train[de, cellTypes_train == x]))))

    colnames(centroidMat) <- unique(cellTypes_train)

    # Constructing the tree using selected tree method
    if (verbose) {
      print("Constructing tree ...")
    }

    cutree_list <- constructTree(centroidMat,
                                 tree = tree,
                                 hopach_kmax = hopach_kmax,
                                 plot = verbose)


  } else {
    cutree_list <- cellType_tree
  }

  if (verbose) {
    print("Training....")
  }




  if (parallel) {
    hierarchyKNNRes <- BiocParallel::bplapply(seq_len(length(selectFeatures)),
                                              function(ft)
                                                hierarchyKNNcor(exprsMat_train,
                                                                cellTypes_train,
                                                                cutree_list,
                                                                feature = selectFeatures[ft],
                                                                topN = topN,
                                                                pSig = pSig,
                                                                verbose = verbose),
                                              BPPARAM = BPPARAM)
    names(hierarchyKNNRes) <- selectFeatures
  }else{
    hierarchyKNNRes <- list()
    for (ft in seq_len(length(selectFeatures))) {

      if (verbose) {
        print(paste("=== selecting features by:", selectFeatures[ft], "===="))
      }

      hierarchyKNNRes[[ft]] <- hierarchyKNNcor(exprsMat_train,
                                               cellTypes_train,
                                               cutree_list,
                                               feature = selectFeatures[ft],
                                               topN = topN,
                                               pSig = pSig,
                                               verbose = verbose)
    }

    names(hierarchyKNNRes) <- selectFeatures
  }


  trainRes <- list(hierarchyKNNRes = hierarchyKNNRes,
                   cutree_list = cutree_list,
                   cellTypes_train = cellTypes_train)


  if (weightsCal) {
    if (verbose) {
      cat("=========  SelfTraining to calculate weight  ================= \n")
    }

    selfTrainRes <- predict_scClassify(exprsMat_test = exprsMat_train,
                                       trainRes  = trainRes,
                                       cellTypes_test = cellTypes_train,
                                       parallel = parallel,
                                       BPPARAM = BPPARAM,
                                       verbose = verbose,
                                       features = selectFeatures,
                                       ...)
    trainRes$selfTrainRes <- selfTrainRes
    trainRes$modelweights <- getTrainWeights(selfTrainRes)
  }



  return(trainRes)
}



# Function to construct tree using HOPACH or HC

constructTree <- function(centroidMat,
                          tree = c("HOPACH", "HC"),
                          hopach_kmax = hopach_kmax,
                          plot= TRUE){


  tree <- match.arg(tree, c("HOPACH", "HC"), several.ok = FALSE)



  if (tree == "HOPACH") {

    res <- runHOPACH(data = t(centroidMat),
                     plot = plot,
                     kmax = hopach_kmax)
    cutree_list <- res$cutree_list

  }else{
    distMat <- as.dist(1 - stats::cor(centroidMat))
    hc <- stats::hclust(distMat, method = "complete")
    # plot(hc)
    cutree_list <- cutree_iterative(hc, depth = 1)
  }

  return(cutree_list)
}


# Cut the Hierarchical tree for each level
# depth here indicates the deep the tree is cut

cutree_iterative <- function(hc, depth = 1){

  height_sort <- sort(hc$height, decreasing = TRUE)
  cutree_list <- list()

  for (i in seq_len(sum(height_sort >=
                        height_sort[round(length(height_sort)*depth)]))) {
    cutree_list[[i]] <- cutree(hc, h = height_sort[i])
  }

  # Last level is distinct number
  cutree_list[[length(cutree_list) + 1]] <- seq_len(length(hc$labels))
  names(cutree_list[[length(cutree_list)]]) <- hc$labels

  return(cutree_list)
}

currentClass <- function(cellTypes, cutree_res){
  cellTypes <- as.character(cellTypes)
  res <- cutree_res[cellTypes]
  return(res)
}

# Function to perform hierarchical feature selection

hierarchyKNNcor <- function(exprsMat,
                            cellTypes,
                            cutree_list,
                            feature = c("limma", "DV", "DD", "chisq", "BI"),
                            topN = 50,
                            pSig = 0.001,
                            verbose= TRUE){
  feature <- match.arg(feature, c("limma", "DV", "DD", "chisq", "BI", "Cepo"))
  numHierchy <- length(cutree_list)
  levelModel <- list()
  levelHVG <- list()
  for (i in seq_len(numHierchy)) {
    #Make sure not all same cluster
    if (length(unique(cutree_list[[i]])) != 1) {

      model <- list()
      hvg <- list()
      class_tmp <- currentClass(cellTypes, cutree_list[[i]])
      names(class_tmp) <- colnames(exprsMat)
      for (j in unique(cutree_list[[i - 1]])) {
        # print(paste("Group", j))
        trainIdx <- which(cellTypes %in% names(cutree_list[[i - 1]])[cutree_list[[i - 1]] == j])
        trainClass <- class_tmp[trainIdx]
        if (length(unique(trainClass)) != 1) {

          hvg[[j]] <- featureSelection(exprsMat[,trainIdx],
                                       trainClass,
                                       feature = feature,
                                       topN = topN,
                                       pSig = pSig
          )

          model[[j]] <- list(train = Matrix::t(exprsMat[na.omit(hvg[[j]]),
                                                        trainIdx,
                                                        drop = FALSE]),
                             y = as.factor(trainClass))

        }else{
          model[[j]] <- "noModel"
        }
      }
      levelHVG[[i]] <- hvg
      levelModel[[i]] <- model


    }
  }
  res <- list(model = levelModel, hvg = levelHVG)
  return(res)
}

#' @importFrom graphics plot abline
#' @importFrom stats uniroot qnorm dnorm coef
#' @importFrom mixtools normalmixEM
#' @importFrom methods slot

KNNcor <- function(corMat,
                   subLevelModel,
                   cutoff_method = c("dynamic", "static"),
                   k = 10,
                   prob_threshold = 0.8,
                   cor_threshold_static = 0.5,
                   cor_threshold_high = 0.7,
                   topLevel = FALSE,
                   verbose = TRUE) {


    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"),
                               several.ok = FALSE)


    # If we use static correlation cutoff
    if (cutoff_method == "static") {
        if (verbose) {
            print("Using static correlation cutoff...")
        }

        # get the KNN labels
        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])

        # get the KNN correlation
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])

        # If the KNN labels with correlation less than the threshood,
        # set the labels as -1
        topKNN <- ifelse(topKNN_cor >= cor_threshold_static, topKNN, -1)


        # Get the predicted results
        predRes <- apply(topKNN, 2, function(x) {
            tab <- table(x)/length(x)
            if (max(tab, na.rm = TRUE) < prob_threshold) {
                0
            }else{
                if (names(tab)[which(tab == max(tab, na.rm = TRUE))] == "-1") {
                    0
                }else{
                    names(tab)[which(tab == max(tab, na.rm = TRUE))]
                }
            }
        })

    }


    # If we use dynamic correlation cutoff based on mixture model
    if (cutoff_method == "dynamic") {

        if (verbose) {
            print("Using static correlation cutoff...")
        }


        # If this is the top level


        unique_y <- levels(subLevelModel$y)
        cor_threshold_tmp <- c()

        for (l in seq_len(length(unique_y))) {


            if ("dgCMatrix" %in% is(corMat)) {
                corMat_vec <- methods::slot(corMat, "x")
            } else {
                corMat_vec <- c(as.matrix(corMat[subLevelModel$y ==
                                                     unique_y[l],]))
            }

            # Fitting the mixture model if it is not unimodal distribution.
            mix_res <- mix_threshold(corMat_vec = corMat_vec,
                                     length_unique_y = length(unique_y),
                                     verbose = verbose)
            dip_test <- mix_res$dip_test
            mixmdl <- mix_res$mixmdl

            # Caculate the threshold for this branch
            if (dip_test$p.value > 0.01) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else if ("try-error" %in% is(mixmdl)) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else{
                # plot(mixmdl,which = 2)
                t_G2 <- getThreshold(mixmdl, verbose = verbose)
                cor_threshold_tmp = c(cor_threshold_tmp, t_G2)
            }


        }
        names(cor_threshold_tmp) <- unique_y

        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])

        # get the threshold
        # the threshold_tmp is order based on the levels of factor.
        #no need to change the character
        topKNN_threshold <- apply(corMat, 2, function(x) {
            num_l <- subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)]
            num_l <- as.numeric(num_l)
            cor_threshold_tmp[num_l]
        })

        topKNN <- ifelse(topKNN_cor >= topKNN_threshold, topKNN, -1)


        predRes <- apply(topKNN, 2, function(x){
            # x <- stats::na.omit(x)
            tab <- table(x)/length(x)
            if (max(tab, na.rm = TRUE) < prob_threshold) {
                0
            }else{
                if (names(tab)[which(tab == max(tab, na.rm = TRUE))] == "-1") {
                    0
                }else{
                    names(tab)[which(tab == max(tab, na.rm = TRUE))]
                }

            }
        })




    }

    return(list(predRes = predRes))
}

#' @importFrom methods slot
WKNNcor <- function(corMat,
                    subLevelModel,
                    cutoff_method = c("dynamic", "static"),
                    k = 10,
                    prob_threshold = 0.8,
                    cor_threshold_static = 0.5,
                    cor_threshold_high = 0.7,
                    topLevel = FALSE,
                    verbose = TRUE){

    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"),
                               several.ok = FALSE)

    if (cutoff_method == "static") {

        if (verbose) {
            print("Using static correlation cutoff...")
        }


        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN <- ifelse(topKNN_cor >= cor_threshold_static, topKNN, -1)

        topKNN_weight <- apply(topKNN_cor, 2, function(x){
            x <- stats::na.omit(x) # add in 20190613
            if (x[1] == x[length(x)]) {
                rep(1, length(x))
            } else{
                (x - x[length(x)])/(x[1] - x[length(x)])
            }
        })


        predRes <- getNumericPredRes(corMat, topKNN_weight,
                                     topKNN, prob_threshold)

    }

    if (cutoff_method == "dynamic") {

        if (verbose) {
            print("Using dynamic correlation cutoff...")
        }


        unique_y <- levels(subLevelModel$y)
        cor_threshold_tmp <- c()
        for (l in seq_len(length(unique_y))) {

            if ("dgCMatrix" %in% is(corMat)) {
                corMat_vec <- methods::slot(corMat, "x")
            } else {
                corMat_vec <- c(as.matrix(corMat[subLevelModel$y ==
                                                     unique_y[l],]))
            }


            # Fitting the mixture model if it is not unimodal distribution.
            mix_res <- mix_threshold(corMat_vec = corMat_vec,
                                     length_unique_y = length(unique_y),
                                     verbose = verbose)
            dip_test <- mix_res$dip_test
            mixmdl <- mix_res$mixmdl

            # Caculate the threshold for this branch
            if (dip_test$p.value > 0.01) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else if ("try-error" %in% is(mixmdl)) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else{
                # plot(mixmdl,which = 2)
                t_G2 <- getThreshold(mixmdl, verbose = verbose)
                cor_threshold_tmp = c(cor_threshold_tmp, t_G2)
            }

        }
        names(cor_threshold_tmp) <- unique_y


        ### calcualte the Weighted KNN
        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_threshold <- apply(corMat, 2, function(x) {
            num_l <- subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)]
            num_l <- as.numeric(num_l)
            cor_threshold_tmp[num_l]
        })



        topKNN <- ifelse(topKNN_cor >= topKNN_threshold, topKNN, -1)

        topKNN_weight <- apply(topKNN_cor, 2, function(x){
            #x <- stats::na.omit(x) # add 20190613
            x[is.na(x)] <- 0
            if (x[1] == x[length(x)]) {
                rep(1, length(x))
            }else{
                (x - x[length(x)])/(x[1] - x[length(x)])
            }
        })

        predRes <- getNumericPredRes(corMat, topKNN_weight,
                                     topKNN, prob_threshold)
    }

    return(list(predRes = predRes))
}





#' @importFrom methods slot
DWKNNcor <- function(corMat,
                     subLevelModel,
                     cutoff_method = c("dynamic", "static"),
                     k = 10,
                     prob_threshold = 0.8,
                     cor_threshold_static = 0.5,
                     cor_threshold_high = 0.7,
                     topLevel = FALSE,
                     verbose = TRUE){

    cutoff_method <- match.arg(cutoff_method, c("dynamic", "static"),
                               several.ok = FALSE)

    if (cutoff_method == "static") {

        if (verbose) {
            print("Using static correlation cutoff...")
        }


        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN <- ifelse(topKNN_cor >= cor_threshold_static, topKNN, -1)



        topKNN_weight <- apply(topKNN_cor, 2, function(x){
            #x <- stats::na.omit(x) # add in 20190613
            x[is.na(x)] <- 0
            if (x[1] == x[length(x)]) {
                rep(1, length(x))
            }else{
                (x - x[length(x)]) *
                    (2 - x[length(x)] - x[1])/((x[1] - x[length(x)]) *
                                                   (2 - x[length(x)] - x))
            }
        })

        predRes <- getNumericPredRes(corMat, topKNN_weight,
                                     topKNN, prob_threshold)

    }

    if (cutoff_method == "dynamic") {

        if (verbose) {
            print("Using dynamic correlation cutoff...")
        }

        unique_y <- levels(subLevelModel$y)
        cor_threshold_tmp <- c()
        for (l in seq_len(length(unique_y))) {

            if ("dgCMatrix" %in% is(corMat)) {
                corMat_vec <- methods::slot(corMat, "x")
            } else {
                corMat_vec <- c(as.matrix(corMat[subLevelModel$y ==
                                                     unique_y[l],]))
            }



            # Fitting the mixture model if it is not unimodal distribution.
            mix_res <- mix_threshold(corMat_vec = corMat_vec,
                                     length_unique_y = length(unique_y),
                                     verbose = verbose)
            dip_test <- mix_res$dip_test
            mixmdl <- mix_res$mixmdl

            # Caculate the threshold for this branch
            if (dip_test$p.value > 0.01) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else if ("try-error" %in% is(mixmdl)) {
                cor_threshold_tmp = c(cor_threshold_tmp, 0)
            }else{
                # plot(mixmdl,which = 2)
                t_G2 <- getThreshold(mixmdl, verbose = verbose)
                cor_threshold_tmp = c(cor_threshold_tmp, t_G2)
            }


        }
        names(cor_threshold_tmp) <- unique_y

        ### calcualte the Weighted KNN
        topKNN <- apply(corMat, 2, function(x)
            subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_cor <- apply(corMat, 2, function(x)
            x[order(x, decreasing = TRUE)][seq_len(k)])
        topKNN_threshold <- apply(corMat, 2, function(x) {
            num_l <- subLevelModel$y[order(x, decreasing = TRUE)][seq_len(k)]
            num_l <- as.numeric(num_l)
            cor_threshold_tmp[num_l]
        })

        topKNN <- ifelse(topKNN_cor >= topKNN_threshold, topKNN, -1)

        topKNN_weight <- apply(topKNN_cor, 2, function(x){
            #x <- stats::na.omit(x) # add in 20190613
            x[is.na(x)] <- 0
            if (x[1] == x[length(x)]) {
                rep(1, length(x))
            }else{
                (x - x[length(x)]) * (2 - x[length(x)] - x[1]) /
                    ((x[1] - x[length(x)]) * (2 - x[length(x)] - x))
            }
        })





        predRes <- getNumericPredRes(corMat, topKNN_weight,
                                     topKNN, prob_threshold)


    }

    return(list(predRes = predRes))
}

# function to fit a mixture model

mix_threshold <- function(corMat_vec, length_unique_y, verbose = verbose) {

    corMat_vec <- thin_cor(corMat_vec, min(10000, length(corMat_vec)))
    corMat_vec <- corMat_vec[corMat_vec != min(corMat_vec)]
    suppressMessages(dip_test <- diptest::dip.test(corMat_vec, B = 10000))
    if (dip_test$p.value <= 0.01) {
        quiet(mixmdl <- try(mixtools::normalmixEM(corMat_vec,
                                                  fast = TRUE,
                                                  maxrestarts = 100,
                                                  k = length_unique_y,
                                                  maxit = 1000,
                                                  mu = c(-0.5, rep(0.5, length_unique_y - 2),  1),
                                                  lambda = c(1/length_unique_y),
                                                  sigma = rep(0.2,
                                                              length_unique_y),
                                                  ECM = TRUE,
                                                  verb = verbose),
                            silent = TRUE))

        if (!"try-error" %in% is(mixmdl)) {
            if (suppressWarnings(min(unique(mixmdl$rho)) == 0)) {
                quiet(mixmdl <- try(mixtools::normalmixEM(corMat_vec,
                                                          fast = TRUE,
                                                          maxrestarts = 100,
                                                          k = length_unique_y,
                                                          maxit = 2000,
                                                          ECM = TRUE,
                                                          verb = verbose),
                                    silent = TRUE))
            }
        }else{
            quiet(mixmdl <- try(mixtools::normalmixEM(corMat_vec, fast = TRUE,
                                                      maxrestarts = 100,
                                                      k = length_unique_y,
                                                      maxit = 2000,
                                                      ECM = TRUE,
                                                      verb = verbose),
                                silent = TRUE))
        }
    } else {
        mixmdl <- NULL
    }

    return(list(mixmdl = mixmdl, dip_test = dip_test))
}



# Function to generate the mixture model
# based on the mixtools normalmixEM results

funMixModel <- function(x, mu1, mu2, sd1, sd2, rho1, rho2) {

    dnorm(x, mean = mu1, sd = sd1) * rho1 -
        dnorm(x, mean = mu2, sd = sd2) * rho2


}




# Function to get the threshold for correlation based on mixture model

getThreshold <- function(mixmdl, verbose = FALSE){

    # if (verbose) {
    #   plot(mixmdl, which = 2)
    # }

    membership <- apply(mixmdl$posterior, 1, which.max)
    m_list <- sort(unique(membership))

    mu_list <- mixmdl$mu
    names(mu_list) <- seq_len(length(mu_list))
    mu_list <- mu_list[m_list]

    if (length(mu_list) > 1) {
        idx1 <- as.numeric(names(mu_list)[order(mu_list)][1])
        idx2 <- as.numeric(names(mu_list)[order(mu_list)][2])

        root <- try(uniroot(funMixModel, interval = c(mixmdl$mu[idx1] -
                                                          mixmdl$sigma[idx1],
                                                      mixmdl$mu[idx2] +
                                                          mixmdl$sigma[idx2]),
                            mu1 = mixmdl$mu[idx1], mu2 = mixmdl$mu[idx2],
                            sd1 = mixmdl$sigma[idx1], sd2 = mixmdl$sigma[idx2],
                            rho1 = mixmdl$lambda[idx1],
                            rho2 = mixmdl$lambda[idx2]),
                    silent = TRUE)


        if (!"try-error" %in%  is(root)) {
            t <- root$root
        }else{
            t <- 0
        }

    }else{
        t <- 0
    }

    return(t)
}



thin_cor <- function(cor_vector, n = 10000) {

    N <- length(cor_vector)

    if (n == N) {
        return(cor_vector)
    }

    new_n <- floor(N/n)

    index <- seq(from = 1, to = N, by = new_n)

    cor_vector <- cor_vector[index]

    return(cor_vector)
}


##

getNumericPredRes <- function(corMat, topKNN_weight, topKNN,
                              prob_threshold) {
    #
    # predRes <- sapply(seq_len(ncol(corMat)),
    #                   function(i){
    #                       vote <- stats::aggregate(topKNN_weight[,i],
    #                                                by = list(topKNN[,i]),
    #                                                sum)
    #                       maxIdx <- which.max(vote$x)
    #                       if (max(vote$x/sum(vote$x)) < prob_threshold) {
    #                           0
    #                       }else{
    #                           if (vote$Group.1[maxIdx] == "-1") {
    #                               0
    #                           }else{
    #                               vote$Group.1[maxIdx]
    #                           }
    #                       }
    #                   })

    predRes <- vapply(seq_len(ncol(corMat)),
                      function(i){
                          vote <- stats::aggregate(topKNN_weight[,i],
                                                   by = list(topKNN[,i]),
                                                   sum)
                          maxIdx <- which.max(vote$x)
                          if (max(vote$x/sum(vote$x)) < prob_threshold) {
                              "0"
                          }else{
                              if (vote$Group.1[maxIdx] == "-1") {
                                  "0"
                              }else{
                                  vote$Group.1[maxIdx]
                              }
                          }
                      }, character(1L))

    names(predRes) <- colnames(corMat)


    return(predRes)
}

#' Fit learning curve for accuracy matrix
#' @param accMat Matrix of accuracy rate where column indicate
#' different sample size
#' @param n Vector indicates the sample size
#' @param auto_initial whether automatical intialise
#' @param a input the parameter a starting point
#' @param b input the parameter a starting point
#' @param c input the parameter a starting point
#' @param d_list range of d
#' @param fitmodel "nls", "nls_mix", "gam"
#' @param plot indicates whether plot or not
#' @param verbose indicates whether verbose or not
#' @return list of results
#' @examples
#' set.seed(2019)
#' n <- seq(20, 10000, 100)
#' accMat <- do.call(cbind, lapply(1:length(n), function(i){
#' tmp_n <- rep(n[i], 50)
#' y <- -2/(tmp_n^0.8) + 0.95 + rnorm(length(tmp_n), 0, 0.02)
#' }))
#' res <- learningCurve(accMat = accMat, n)
#' N <- getN(res, acc = 0.9)
#'
#' @author Yingxin Lin
#' @import ggplot2
#' @importFrom minpack.lm nlsLM
#' @importFrom stats quantile coef predict
#' @importFrom mgcv gam
#' @export

learningCurve <- function(accMat, n, auto_initial = TRUE,
                          a = NULL, b = NULL, c = NULL, d_list = NULL,
                          fitmodel = c("nls", "nls_mix", "gam"),
                          plot = TRUE, verbose = TRUE){

    ## TODO: to make several ok TRUE
    fitmodel <- match.arg(fitmodel, c("nls", "nls_mix", "gam"),
                          several.ok = FALSE)



    if (any(c("matrix", "data.frame") %in% is(accMat))) {
        if (ncol(accMat) != length(n)) {
            stop("Number of column doesn't match with the length of n")
        }
        dat <- data.frame(y = colMeans(accMat),
                          y_25 = apply(accMat, 2,
                                       function(x) quantile(x, 0.25)),
                          y_75 = apply(accMat, 2,
                                       function(x) quantile(x, 0.75))
        )
    }

    if ( "list" %in% is(accMat)) {
        if (length(accMat) != length(n)) {
            stop("Number of column doesn't match with the length of n")
        }
        dat <- data.frame(y = unlist(lapply(accMat, mean)),
                          y_25 = unlist(lapply(accMat,
                                               function(x) quantile(x, 0.25))),
                          y_75 = unlist(lapply(accMat,
                                               function(x) quantile(x, 0.75))))
    }



    # Fitting model
    model <- list()
    for (i in seq(ncol(dat))) {
        if (fitmodel == "nls_mix") {
            model[[i]] <- fitLC_mixture(dat[,i], n,
                                        b = b, d_list = d_list,
                                        verbose = verbose)

        } else if (fitmodel == "gam") {
            model[[i]] <- mgcv::gam(acc ~ s(n, bs = "cr"),
                                    data = data.frame(acc = dat[,i], n = n),
                                    method = "REML")
        }

        else{
            model[[i]] <- fitLC(dat[,i], n, auto_initial = auto_initial,
                                a = a, b = b, c = c,
                                verbose = verbose, b_max = 1)
            if (i == 1) {
                b_max <- coef(model[[i]])["b"]
            }
        }

    }


    # Get fitted value
    dat$n <- n
    new_n <- data.frame(n = seq(min(n), max(n), by = 0.1))
    fit <- lapply(model, function(x) predict(x, newdata = new_n))

    names(model) <- names(fit) <- c("mean", "quantile_25", "quantile_75")
    fit[["n"]] <- new_n$n
    dat_fit <- data.frame(do.call(cbind, fit))

    if (plot) {


        cols <- c("Mean" = "#c8133b","Quantile25/75" = "#ea8783")
        g <-  ggplot2::ggplot(dat, ggplot2::aes(x = n, y = y))  +
            xlab("N") + ylab("Accuracy Rate") +
            ggplot2::geom_point(alpha = 0.7) +
            ggplot2::geom_line(data = dat_fit, aes(x = n,
                                                   y = mean,
                                                   color = "Mean"),
                               linetype = "solid", size = 1) +
            ggplot2::geom_line(data = dat_fit, aes(x = n, y = quantile_25,
                                                   color = "Quantile25/75"),
                               linetype = "dashed") +
            ggplot2::geom_line(data = dat_fit, aes(x = n, y = quantile_75,
                                                   color = "Quantile25/75"),
                               linetype = "dashed") +
            ggplot2::scale_color_manual(values = cols) +
            ggplot2::theme_bw() +
            ggplot2::theme(legend.position = "bottom")

        return(list(model = model, fit = fit, plot = g))

    }else{
        return(list(model = model, fit = fit))
    }
}


init_fit <- function(acc, n){
    fit <- lm(log(n) ~ log(max(acc) + 0.001 - acc))

    c <- -coef(fit)[2]
    a <- -exp(coef(fit)[1])
    return(list(a = a, c = c))
}

# Function to fit the learning curve by inverse power function

fitLC <- function(acc, n, auto_initial = TRUE,
                  a = NULL, b = NULL, c = NULL, verbose = TRUE,
                  b_max = NULL){
    dat_train <- data.frame(n = n, acc = acc)

    if (is.null(b)) {
        b = 0.9
    }

    if (is.null(b_max)) {
        b_max = 1
    }

    if (auto_initial) {
        init <- init_fit(acc, n)

        lower_bound <- c(-max(max(dat_train$acc) + 0.01, 1)*(10^init$c),
                         -Inf, 0)

        learning_curve <-  try({minpack.lm::nlsLM(acc ~ I(1 / n^(c) * a) + b,
                                                  data = dat_train,
                                                  start = list(a = init$a,
                                                               c = init$c,
                                                               b = b),
                                                  upper = c(10, Inf, b_max),
                                                  lower = lower_bound
        )
        }, silent = TRUE)


    }else if (!is.null(a) & !is.null(b) & !is.null(c)) {
        # For the case that user supplies starting point
        learning_curve <- stats::nls(acc ~ I(1 / n^(c) * a) + b,
                                     data = dat_train,
                                     start = list(a = a, c = c, b = b))
    }else{
        # this starting point may not work all the time
        learning_curve <- stats::nls(acc ~ I(1 / n^(c) * a) + b,
                                     data = dat_train,
                                     start = list(a = -10, c = 1, b = 1))
    }

    para <- coef(learning_curve)
    names(para) <- c("a1", "c1", "b1")

    if (verbose) {
        print(summary(learning_curve))
    }

    learning_curve$para <- para
    learning_curve$fit_model <- "nls"

    return(learning_curve)
}




fitLC_mixture <- function(acc, n,  b = NULL,
                          d_list = NULL,
                          verbose = TRUE){
    dat_train <- data.frame(n = n, acc = acc)

    if (is.null(b)) {
        b = 0.9
    }

    if (is.null(d_list)) {
        d_list <- seq(min(n) + 10, round(max(n)/2), 10)
    }

    init <- init_fit(acc, n)

    rss <- c()
    for (i in seq_len(length(d_list))) {
        d <- d_list[i]
        learning_curve <-  try({minpack.lm::nlsLM(acc ~ H1(d - n) *
                                                      I(1 / n^(c) * a) +
                                                      H(n - d)  *
                                                      I(1 / n^(c1) *
                                                            a*d^(c1 - c))  + b,
                                                  data = dat_train,
                                                  start = list(a = init$a,
                                                               c = init$c,
                                                               b = b,
                                                               c1 = init$c),
                                                  upper = c(Inf, Inf, 1, Inf),
                                                  lower = c(-Inf, -Inf, 0,
                                                            -Inf),
                                                  control = list(maxiter =
                                                                     1000))},
                               silent = TRUE)

        if (!"try-error" %in% is(learning_curve)) {
            rss <- c(rss, summary(learning_curve)$sigma)
        }else{
            rss <- c(rss, Inf)
        }

    }


    d_opt <- d_list[which.min(rss)]

    learning_curve <-  minpack.lm::nlsLM(acc ~ H1(d_opt - n) *
                                             I(1 / n^(c) * a) +
                                             H(n - d_opt)  *
                                             I(1 / n^(c1) * a*d_opt^(c1-c))  +
                                             b,
                                         data = dat_train,
                                         start = list(a = init$a, c = init$c,
                                                      b = 0.95,
                                                      c1 = init$c),
                                         upper = c(Inf, Inf, 1, Inf),
                                         lower = c(-Inf, -Inf, 0, -Inf),
                                         control = list(maxiter = 1000)

    )

    para <- coef(learning_curve)
    para <- c(para, para[1]*d_opt^(para[4] - para[2]), d_opt)
    names(para) <- c("a1", "c1", "b1", "c2", "a2", "d")

    if (verbose) {
        print(summary(learning_curve))
    }
    learning_curve$para <- para
    learning_curve$fit_model <- "nls_mix"
    return(learning_curve)
}


#' Function to get the required N given by the accuracy and
#' the learning curve model
#' @param res model results returned by \code{learning_curve} function
#' @param acc accuracy that are quired
#' @return sample size that are required
#'
#' @examples
#' set.seed(2019)
#' n <- seq(20, 10000, 100)
#' accMat <- do.call(cbind, lapply(1:length(n), function(i){
#' tmp_n <- rep(n[i], 50)
#' y <- -2/(tmp_n^0.8) + 0.95 + rnorm(length(tmp_n), 0, 0.02)
#' }))
#' res <- learningCurve(accMat = accMat, n)
#' N <- getN(res, acc = 0.9)
#'
#' @export

getN <- function(res, acc = 0.9) {
    para <- coef(res$model$mean)
    names(para) <- c("a", "c", "b")
    if (acc > para["b"]) {
        stop("Required accuracy is too high to achieve :(")
    }
    N <- round(exp(1/para["c"]*log(para["a"]/(acc - para["b"]))))
    names(N) <- NULL

    return(N)
}

H <- function(x) as.numeric(x > 0)
H1 <- function(x) as.numeric(x >= 0)

---
title: "Performing scClassify using pretrained model"
author:
- name: Yingxin Lin
  affiliation: School of Mathematics and Statistics, The University of Sydney, Australia
date: "`r BiocStyle::doc_date()`"
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
vignette: >
  %\VignetteIndexEntry{pretrainedModel}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
  
  
  
  
```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  warning = FALSE,
  message = FALSE,
  comment = "#>"
)
```



# Introduction

A common application of single-cell RNA sequencing (RNA-seq) data is 
to identify discrete cell types. To take advantage of the large collection 
of well-annotated scRNA-seq datasets, `scClassify` package implements 
a set of methods to perform accurate cell type classification based on 
*ensemble learning* and *sample size calculation*. 

This vignette will provide an example showing how users can use a pretrained 
model of scClassify to predict cell types. A pretrained model is a 
`scClassifyTrainModel` object  returned by `train_scClassify()`. 
A list of pretrained model can be found in
https://sydneybiox.github.io/scClassify/index.html.


First,  install `scClassify`, install `BiocManager` and use 
`BiocManager::install` to install `scClassify` package.

```{r eval = FALSE}
# installation of scClassify
if (!requireNamespace("BiocManager", quietly = TRUE)) {
  install.packages("BiocManager")
}
BiocManager::install("scClassify")
```





# Setting up the data

We assume that you have *log-transformed* (size-factor normalized) matrices as 
query datasets, where each row refers to a gene and each column a cell. 
For demonstration purposes, we will take a subset of single-cell pancreas 
datasets from one independent study (Wang et al.).


```{r setup}
library(scClassify)
data("scClassify_example")
wang_cellTypes <- scClassify_example$wang_cellTypes
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
exprsMat_wang_subset <- as(exprsMat_wang_subset, "dgCMatrix")
```

Here, we load our pretrained model using a subset of the Xin et al. 
human pancreas dataset as our reference data.

First, let us check basic information relating to our pretrained model. 

```{r}
data("trainClassExample_xin")
trainClassExample_xin
```

In this pretrained model, we have selected the genes based on Differential 
Expression using limma. To check the genes that are available 
in the pretrained model: 


```{r}
features(trainClassExample_xin)
```


We can also visualise the cell type tree of the reference data.

```{r}
plotCellTypeTree(cellTypeTree(trainClassExample_xin))
```

# Running scClassify

Next, we perform `predict_scClassify` with our pretrained model 
`trainRes = trainClassExample` to predict the cell types of our 
query data matrix `exprsMat_wang_subset_sparse`. Here, 
we used `pearson` and `spearman` as similarity metrics.

```{r}
pred_res <- predict_scClassify(exprsMat_test = exprsMat_wang_subset,
                               trainRes = trainClassExample_xin,
                               cellTypes_test = wang_cellTypes,
                               algorithm = "WKNN",
                               features = c("limma"),
                               similarity = c("pearson", "spearman"),
                               prob_threshold = 0.7,
                               verbose = TRUE)
```

Noted that the `cellType_test` is not a required input. 
For datasets with unknown labels, users can simply leave it 
as `cellType_test = NULL`.



Prediction results for pearson as the similarity metric:


```{r}
table(pred_res$pearson_WKNN_limma$predRes, wang_cellTypes)
```

Prediction results for spearman as the similarity metric:

```{r}
table(pred_res$spearman_WKNN_limma$predRes, wang_cellTypes)
```




# Session Info

```{r}
sessionInfo()
```

---
title: "scClassify Model Building and Prediction"
author:
- name: Yingxin Lin
  affiliation: 
  - School of Mathematics and Statistics, The University of Sydney, Australia
  - Charles Perkins Centre, The University of Sydney, Australia
date: "`r BiocStyle::doc_date()`"
output:
  BiocStyle::html_document:
    toc: true
    toc_float: true
vignette: >
  %\VignetteIndexEntry{scClassify}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---
  
  
```{r, include = FALSE}
knitr::opts_chunk$set(
    collapse = TRUE,
    warning = FALSE,
    message = FALSE,
    comment = "#>"
)
```





# Introduction

A common application of single-cell RNA sequencing (RNA-seq) data is 
to identify discrete cell types. To take advantage of the large collection 
of well-annotated scRNA-seq datasets, `scClassify` package implements 
a set of methods to perform accurate cell type classification based on 
*ensemble learning* and *sample size calculation*. 
This vignette demonstrates the usage of `scClassify`, 
providing a pithy description of each method with workable examples. 

First, install `scClassify` via `BiocManager`.

```{r eval = FALSE}
# installation of scClassify
if (!requireNamespace("BiocManager", quietly = TRUE)) {
    install.packages("BiocManager")
}
BiocManager::install("scClassify")
```

# Setting up the data

We assume that you have *log-transformed* (size-factor normalized) matrices 
where each row is a gene and each column a cell for a reference dataset 
and a query dataset. For demonstration purposes, we will take a subset of 
single-cell pancreas datasets from two independent studies 
(Wang et al., and Xin et al.).


```{r setup}
library("scClassify")
data("scClassify_example")
xin_cellTypes <- scClassify_example$xin_cellTypes
exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
wang_cellTypes <- scClassify_example$wang_cellTypes
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
exprsMat_xin_subset <- as(exprsMat_xin_subset, "dgCMatrix")
exprsMat_wang_subset <- as(exprsMat_wang_subset, "dgCMatrix")
```


The original cell type annotations and compositions of the example datasets 
can be easily accessed as shown below.


```{r}
table(xin_cellTypes)
table(wang_cellTypes)
```

We can see that Xin et al. data only have 4 cell types, 
while Wang et al. has 7 cell types.


# scClassify



## Non-ensemble scClassify

We first perform non-ensemble `scClassify` by using Xin et al. 
as our reference dataset and Wang et al. data as ur query data. 
We use `WKNN` as the KNN algorithm, `DE` (differential expression genes) 
as the gene selection method, and lastly `pearson` as 
the similarity calculation method.

```{r}
scClassify_res <- scClassify(exprsMat_train = exprsMat_xin_subset,
                             cellTypes_train = xin_cellTypes,
                             exprsMat_test = list(wang = exprsMat_wang_subset),
                             cellTypes_test = list(wang = wang_cellTypes),
                             tree = "HOPACH",
                             algorithm = "WKNN",
                             selectFeatures = c("limma"),
                             similarity = c("pearson"),
                             returnList = FALSE,
                             verbose = FALSE)

```

We can check the cell type tree generated by the reference data:

```{r warning=FALSE}
scClassify_res$trainRes
plotCellTypeTree(cellTypeTree(scClassify_res$trainRes))
```

Noted that `scClassify_res$trainRes` is a `scClassifyTrainModel` class.


Check the prediction results.

```{r}
table(scClassify_res$testRes$wang$pearson_WKNN_limma$predRes, wang_cellTypes)
```




## Ensemble Classify


We next perform ensemble `scClassify` by using Xin et al. 
as our reference dataset and Wang et al. data as our query data. 
We use `WKNN` as the KNN algorithm, `DE` as the gene selection method, 
and `pearson` and `spearman` as the similarity calculation methods. 
Thus, we will generate two combinations of gene selection models and 
similarity metrics as training classifiers: 

1. `WKNN` + `DE` + `pearson`
2. `WKNN` + `DE` + `spearman`

Here, we will weight these two classifiers equally by setting 
`weighted_ensemble = FALSE`. By default this is set as `TRUE`, 
so each base classifier will be weighted by the accuracy rates trained 
in the reference data.

```{r}
scClassify_res_ensemble <- scClassify(exprsMat_train = exprsMat_xin_subset,
                                      cellTypes_train = xin_cellTypes,
                                      exprsMat_test = list(wang = exprsMat_wang_subset),
                                      cellTypes_test = list(wang = wang_cellTypes),
                                      tree = "HOPACH",
                                      algorithm = "WKNN",
                                      selectFeatures = c("limma"),
                                      similarity = c("pearson", "cosine"),
                                      weighted_ensemble = FALSE,
                                      returnList = FALSE,
                                      verbose = FALSE)

```

We can compare the two base classifiers predictions as below.

```{r}
table(scClassify_res_ensemble$testRes$wang$pearson_WKNN_limma$predRes,
      scClassify_res_ensemble$testRes$wang$cosine_WKNN_limma$predRes)
```


Now, check the final ensemble results:


```{r}
table(scClassify_res_ensemble$testRes$wang$ensembleRes$cellTypes, 
      wang_cellTypes)
```




# Train your own model

You can also train your own model `scClassifyTrainModel` using 
`train_scClassify()`. Note that by setting `weightsCal = TRUE`,
we will calculate the training error of the reference data as 
the weights for the individual classifiers.

Here, we illustrate the training function with gene selection methods 
based on differential expression ("limma") and biomodal distribution ("BI").

```{r}
trainClass <- train_scClassify(exprsMat_train = exprsMat_xin_subset,
                               cellTypes_train = xin_cellTypes,
                               selectFeatures = c("limma", "BI"),
                               returnList = FALSE
)
```


```{r}
trainClass
```



# Session Info

```{r}
sessionInfo()
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassify.R
\name{scClassify}
\alias{scClassify}
\title{Train and test scClassify model}
\usage{
scClassify(
  exprsMat_train = NULL,
  cellTypes_train = NULL,
  exprsMat_test = NULL,
  cellTypes_test = NULL,
  tree = "HOPACH",
  algorithm = "WKNN",
  selectFeatures = "limma",
  similarity = "pearson",
  cutoff_method = c("dynamic", "static"),
  weighted_ensemble = FALSE,
  weights = NULL,
  weighted_jointClassification = TRUE,
  cellType_tree = NULL,
  k = 10,
  topN = 50,
  hopach_kmax = 5,
  pSig = 0.01,
  prob_threshold = 0.7,
  cor_threshold_static = 0.5,
  cor_threshold_high = 0.7,
  returnList = TRUE,
  parallel = FALSE,
  BPPARAM = BiocParallel::SerialParam(),
  verbose = FALSE
)
}
\arguments{
\item{exprsMat_train}{A matrix of log-transformed expression matrix of reference dataset}

\item{cellTypes_train}{A vector of cell types of reference dataset}

\item{exprsMat_test}{A list or a matrix indicates the expression matrices of the query datasets}

\item{cellTypes_test}{A list or a vector indicates cell types of the query datasets (Optional).}

\item{tree}{A vector indicates the method to build hierarchical tree, set as "HOPACH" by default.
This should be one of "HOPACH" and "HC" (using hclust).}

\item{algorithm}{A vector indicates the KNN method that are used, set as
"WKNN" by default. Thisshould be one or more of "WKNN", "KNN", "DWKNN".}

\item{selectFeatures}{A vector indicates the gene selection method, set as "limma" by default.
This should be one or more of "limma", "DV", "DD", "chisq", "BI" and "Cepo".}

\item{similarity}{A vector indicates the similarity measure that are used,
set as "pearson" by default. This should be one or more of "pearson",
"spearman", "cosine", "jaccard", kendall", "binomial", "weighted_rank","manhattan"}

\item{cutoff_method}{A vector indicates the method to cutoff the correlation distribution.
Set as "dynamic" by default.}

\item{weighted_ensemble}{A logical input indicates in ensemble learning,
whether the results is combined by a weighted score for each base classifier.}

\item{weights}{A vector indicates the weights for ensemble}

\item{weighted_jointClassification}{A logical input indicates in joint classification
using multiple training datasets,
whether the results is combined by a weighted score for each training model.}

\item{cellType_tree}{A list indicates the cell type tree provided by user.
(By default, it is NULL) (Only for one training data input)}

\item{k}{An integer indicates the number of neighbour}

\item{topN}{An integer indicates the top number of features that are selected}

\item{hopach_kmax}{An integer between 1 and 9 specifying the maximum number of
children at each node in the HOPACH tree.}

\item{pSig}{A numeric indicates the cutoff of pvalue for features}

\item{prob_threshold}{A numeric indicates the probability threshold for KNN/WKNN/DWKNN.}

\item{cor_threshold_static}{A numeric indicates the static correlation threshold.}

\item{cor_threshold_high}{A numeric indicates the highest correlation threshold}

\item{returnList}{A logical input indicates whether the output will be class of list}

\item{parallel}{A logical input indicates whether running in paralllel or not}

\item{BPPARAM}{A \code{BiocParallelParam} class object
from the \code{BiocParallel} package is used. Default is SerialParam().}

\item{verbose}{A logical input indicates whether the intermediate steps will be printed}
}
\value{
A list of the results, including testRes storing the results of the testing information,
and trainRes storing the training model inforamtion.
}
\description{
Train and test scClassify model
}
\examples{

data("scClassify_example")
xin_cellTypes <- scClassify_example$xin_cellTypes
exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
wang_cellTypes <- scClassify_example$wang_cellTypes
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset

scClassify_res <- scClassify(exprsMat_train = exprsMat_xin_subset,
cellTypes_train = xin_cellTypes,
exprsMat_test = list(wang = exprsMat_wang_subset),
cellTypes_test = list(wang = wang_cellTypes),
tree = "HOPACH",
algorithm = "WKNN",
selectFeatures = c("limma"),
similarity = c("pearson"),
returnList = FALSE,
verbose = FALSE)

}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{scClassify_example}
\alias{scClassify_example}
\title{Example data used in scClassify package}
\format{
An object of class \code{list} of length 4.
}
\source{
Wang YJ, Schug J, Won K-J, Liu C, Naji A, Avrahami D, Golson ML &
Kaestner KH (2016) Single cell transcriptomics of the human endocrine
pancreas. Diabetes: db160405

Xin Y, Kim J, Okamoto H, Ni M, Wei Y, Adler C, Murphy AJ, Yancopoulos GD,
Lin C & Gromada J (2016) RNA sequencing of single human islet cells reveals
type 2 diabetes genes. Cell Metab. 24: 608–615
}
\usage{
data(scClassify_example, package = 'scClassify')
}
\description{
A list includes expression matrix and cell type of subsets of
wang et al., xin et al.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\docType{class}
\name{scClassifyTrainModel-class}
\alias{scClassifyTrainModel-class}
\title{An S4 class to stored training model for scClassify}
\description{
An S4 class to stored training model for scClassify
}
\section{Slots}{

\describe{
\item{\code{name}}{Name of the training dataset}

\item{\code{cellTypeTrain}}{A vector of cell type in training dataset}

\item{\code{cellTypeTree}}{A list indicate a cell type tree}

\item{\code{features}}{A vector of character indicates the
features that are trained for this data}

\item{\code{model}}{A list stored the training model,
including the features that are selected
and the cell expression matrix that are used for training}

\item{\code{modelweights}}{A vector of numeric indicates the weights of each model}

\item{\code{metaData}}{A DataFrame stored meta data of training model}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{name}
\alias{name}
\alias{name,scClassifyTrainModel-method}
\title{Accessors of name for scClassifyTrainModel}
\usage{
name(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
name of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
name(trainClassExample_xin)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{.scClassifyTrainModel}
\alias{.scClassifyTrainModel}
\title{The scClassifyTrainModel class}
\usage{
.scClassifyTrainModel(
  name,
  cellTypeTree,
  cellTypeTrain,
  features,
  model,
  modelweights,
  metaData
)
}
\arguments{
\item{name}{Name of the training dataset}

\item{cellTypeTree}{A list indicate a cell type tree}

\item{cellTypeTrain}{A vector of cell type in training dataset}

\item{features}{A vector of character indicates the
features that are trained for this data}

\item{model}{A list stored the training model,
including the features that are selected
and the cell expression matrix that are used for training}

\item{modelweights}{A vector of numeric indicates the weights of each model}

\item{metaData}{A DataFrame stored meta data of training model}
}
\value{
A scClassifyTrainModel object
}
\description{
The scClassifyTrainModel class is designed to stored
training model for scClassify
}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/train_scClassify.R
\name{train_scClassify}
\alias{train_scClassify}
\title{Training scClassify model}
\usage{
train_scClassify(
  exprsMat_train,
  cellTypes_train,
  tree = "HOPACH",
  selectFeatures = "limma",
  topN = 50,
  hopach_kmax = 5,
  pSig = 0.05,
  cellType_tree = NULL,
  weightsCal = FALSE,
  parallel = FALSE,
  BPPARAM = BiocParallel::SerialParam(),
  verbose = TRUE,
  returnList = TRUE,
  ...
)
}
\arguments{
\item{exprsMat_train}{A matrix of log-transformed expression matrix of reference dataset}

\item{cellTypes_train}{A vector of cell types of reference dataset}

\item{tree}{A vector indicates the method to build hierarchical tree,
set as "HOPACH" by default.
This should be one of "HOPACH" and "HC" (using stats::hclust).}

\item{selectFeatures}{A vector indicates the gene selection method,
set as "limma" by default.
This should be one or more of "limma", "DV", "DD", "chisq", "BI", "Cepo".}

\item{topN}{An integer indicates the top number of features that are selected}

\item{hopach_kmax}{An integer between 1 and 9 specifying the maximum number of
children at each node in the HOPACH tree.}

\item{pSig}{A numeric indicates the cutoff of pvalue for features}

\item{cellType_tree}{A list indicates the cell type tree provided by user.
(By default, it is NULL)}

\item{weightsCal}{A logical input indicates whether we need to
calculate the weights for the model.}

\item{parallel}{A logical input indicates whether the algorihms will run in parallel}

\item{BPPARAM}{A \code{BiocParallelParam} class object
from the \code{BiocParallel} package is used. Default is SerialParam().}

\item{verbose}{A logical input indicates whether the intermediate steps will be printed}

\item{returnList}{A logical input indicates whether the output will be class of list}

\item{...}{Other input for predict_scClassify for the case when weights calculation
of the pretrained model is performed}
}
\value{
list of results or an object of \code{scClassifyTrainModel}
}
\description{
Training scClassify model
}
\examples{
data("scClassify_example")
xin_cellTypes <- scClassify_example$xin_cellTypes
exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset
trainClass <- train_scClassify(exprsMat_train = exprsMat_xin_subset,
cellTypes_train = xin_cellTypes,
selectFeatures = c("limma", "BI"),
returnList = FALSE
)

}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learningCurve.R
\name{learningCurve}
\alias{learningCurve}
\title{Fit learning curve for accuracy matrix}
\usage{
learningCurve(
  accMat,
  n,
  auto_initial = TRUE,
  a = NULL,
  b = NULL,
  c = NULL,
  d_list = NULL,
  fitmodel = c("nls", "nls_mix", "gam"),
  plot = TRUE,
  verbose = TRUE
)
}
\arguments{
\item{accMat}{Matrix of accuracy rate where column indicate
different sample size}

\item{n}{Vector indicates the sample size}

\item{auto_initial}{whether automatical intialise}

\item{a}{input the parameter a starting point}

\item{b}{input the parameter a starting point}

\item{c}{input the parameter a starting point}

\item{d_list}{range of d}

\item{fitmodel}{"nls", "nls_mix", "gam"}

\item{plot}{indicates whether plot or not}

\item{verbose}{indicates whether verbose or not}
}
\value{
list of results
}
\description{
Fit learning curve for accuracy matrix
}
\examples{
set.seed(2019)
n <- seq(20, 10000, 100)
accMat <- do.call(cbind, lapply(1:length(n), function(i){
tmp_n <- rep(n[i], 50)
y <- -2/(tmp_n^0.8) + 0.95 + rnorm(length(tmp_n), 0, 0.02)
}))
res <- learningCurve(accMat = accMat, n)
N <- getN(res, acc = 0.9)

}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{modelweights}
\alias{modelweights}
\alias{modelweights,scClassifyTrainModel-method}
\title{Accessors of modelweights for scClassifyTrainModel}
\usage{
modelweights(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
modelweights of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
modelweights(trainClassExample_xin)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sampleSizeCal.R
\name{runSampleCal}
\alias{runSampleCal}
\title{Run sample size calculation for pilot data for reference dataset}
\usage{
runSampleCal(
  exprsMat,
  cellTypes,
  n_list = c(20, 40, 60, 80, 100, seq(200, 500, 100)),
  num_repeat = 20,
  level = NULL,
  cellType_tree = NULL,
  BPPARAM = BiocParallel::SerialParam(),
  subset_test = FALSE,
  num_test = NULL,
  ...
)
}
\arguments{
\item{exprsMat}{A matrix of expression matrix of pilot dataset
(log-transformed, or normalised)}

\item{cellTypes}{A vector of cell types of pilot dataset}

\item{n_list}{A vector of integer indicates the sample size to run.}

\item{num_repeat}{An integer indicates the number of run for
each sample size will be repeated.}

\item{level}{An integer indicates the accuracy rate is calculate
based on the n-th level from top of cell type tree.
If it is NULL (by default), it will be the bottom of the cell type tree.
It can not be larger than the total number of levels of the tree.}

\item{cellType_tree}{A list indicates the cell type tree (optional),
if it is NULL, the accuracy rate is calculate
based on the provided cellTypes.}

\item{BPPARAM}{A \code{BiocParallelParam} class object
from the \code{BiocParallel} package is used. Default is SerialParam().}

\item{subset_test}{A ogical input indicates whether we used a subset of data
(fixed number for each sample size)
to test instead of all remaining data. By default, it is FALSE.}

\item{num_test}{An integer indicates the size of the test data.}

\item{...}{other parameter from scClassify}
}
\value{
A matrix of accuracy matrix, where columns corresponding to different
sample sizes, rows corresponding to the number of repetation.
}
\description{
Run sample size calculation for pilot data for reference dataset
}
\examples{
data("scClassify_example")
xin_cellTypes <- scClassify_example$xin_cellTypes
exprsMat_xin_subset <- scClassify_example$exprsMat_xin_subset

exprsMat_xin_subset <- as(exprsMat_xin_subset, "dgCMatrix")
set.seed(2019)
accMat <- runSampleCal(exprsMat_xin_subset,
xin_cellTypes,
n_list = seq(20, 100, 20),
num_repeat = 5, BPPARAM = BiocParallel::SerialParam())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/predict_scClassify.R
\name{predict_scClassifyJoint}
\alias{predict_scClassifyJoint}
\title{Testing scClassify model (joint training)}
\usage{
predict_scClassifyJoint(
  exprsMat_test,
  trainRes,
  cellTypes_test = NULL,
  k = 10,
  prob_threshold = 0.7,
  cor_threshold_static = 0.5,
  cor_threshold_high = 0.7,
  features = "limma",
  algorithm = "WKNN",
  similarity = "pearson",
  cutoff_method = c("dynamic", "static"),
  parallel = FALSE,
  BPPARAM = BiocParallel::SerialParam(),
  verbose = FALSE
)
}
\arguments{
\item{exprsMat_test}{A list or a matrix indicates the expression matrices of the testing datasets}

\item{trainRes}{A `scClassifyTrainModel` or a `list` indicates scClassify training model}

\item{cellTypes_test}{A list or a vector indicates cell types of the testing datasets (Optional).}

\item{k}{An integer indicates the number of neighbour}

\item{prob_threshold}{A numeric indicates the probability threshold for KNN/WKNN/DWKNN.}

\item{cor_threshold_static}{A numeric indicates the static correlation threshold.}

\item{cor_threshold_high}{A numeric indicates the highest correlation threshold}

\item{features}{A vector indicates the method to select features, set as "limma" by default.
This should be one or more of "limma", "DV", "DD", "chisq", "BI".}

\item{algorithm}{A vector indicates the KNN method that are used, set as "WKNN" by default.
This should be one or more of "WKNN", "KNN", "DWKNN".}

\item{similarity}{A vector indicates the similarity measure that are used,
set as "pearson" by default.
This should be one or more of "pearson",  "spearman", "cosine", "jaccard", "kendall",
"binomial", "weighted_rank","manhattan"}

\item{cutoff_method}{A vector indicates the method to cutoff the correlation distribution.
Set as "dynamic" by default.}

\item{parallel}{A logical input indicates whether running in paralllel or not}

\item{BPPARAM}{A \code{BiocParallelParam} class object
from the \code{BiocParallel} package is used. Default is SerialParam().}

\item{verbose}{A logical input indicates whether the intermediate steps will be printed}
}
\value{
list of results
}
\description{
Testing scClassify model (joint training)
}
\examples{
data("scClassify_example")
wang_cellTypes <- scClassify_example$wang_cellTypes
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
data("trainClassExample_xin")
data("trainClassExample_wang")

trainClassExampleJoint <- scClassifyTrainModelList(trainClassExample_wang,
trainClassExample_xin)

pred_res_joint <- predict_scClassifyJoint(exprsMat_test = exprsMat_wang_subset,
trainRes = trainClassExampleJoint,
cellTypes_test = wang_cellTypes,
algorithm = "WKNN",
features = c("limma"),
similarity = c("pearson"),
prob_threshold = 0.7,
verbose = FALSE)

table(pred_res_joint$jointRes$cellTypes, wang_cellTypes)

}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runHOPACH.R
\name{runHOPACH}
\alias{runHOPACH}
\title{Create HOPACH tree}
\usage{
runHOPACH(data, plot = TRUE, kmax = 5)
}
\arguments{
\item{data}{A matrix of average expression matrix
(each row indicates the gene, each column indicates the cell type)}

\item{plot}{Indicate whether plot or not}

\item{kmax}{Integer between 1 and 9 specifying the maximum number of children
at each node in the tree.}
}
\value{
Return a \code{list} where
\itemize{
\item{cutree_list}: A list indicates the hierarchical cell type tree
\item{plot}: A \code{ggplot} visualise the cell type tree
}
}
\description{
A function generating HOPACH tree using the average expression matrix for
each cell type.
}
\examples{

data("scClassify_example")
wang_cellTypes <- factor(scClassify_example$wang_cellTypes)
exprsMat_wang_subset <- scClassify_example$exprsMat_wang_subset
avgMat_wang <- apply(exprsMat_wang_subset, 1, function(x)
aggregate(x, list(wang_cellTypes), mean)$x)
rownames(avgMat_wang) <- levels(wang_cellTypes)
res_hopach <- runHOPACH(avgMat_wang)
res_hopach$plot

}
\references{
van der Laan, M. J. and Pollard, K. S. (2003)
‘A new algorithm for hybrid hierarchical clustering with
visualization and the bootstrap’,
Journal of Statistical Planning and Inference.
doi: 10.1016/S0378-3758(02)00388-9.
}
\author{
Yingxin Lin
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{cellTypeTree}
\alias{cellTypeTree}
\alias{cellTypeTree,scClassifyTrainModel-method}
\title{Accessors of cellTypeTree for scClassifyTrainModel}
\usage{
cellTypeTree(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
cellTypeTree of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
cellTypeTree(trainClassExample_xin)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/runHOPACH.R
\name{plotCellTypeTree}
\alias{plotCellTypeTree}
\title{To plot cell type tree}
\usage{
plotCellTypeTree(cutree_list, group_level = NULL)
}
\arguments{
\item{cutree_list}{A list indicates the hierarchical cell type tree}

\item{group_level}{Indicate whether plot or not}
}
\value{
A ggplot object visualising the HOPACH tree
}
\description{
To plot cell type tree
}
\examples{

data("trainClassExample_xin")

plotCellTypeTree(cellTypeTree(trainClassExample_xin))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{trainClassExample_wang}
\alias{trainClassExample_wang}
\title{Subset of pretrained model of Wang et al.}
\format{
An object of class \code{scClassifyTrainModel} of length 1.
}
\source{
Wang YJ, Schug J, Won K-J, Liu C, Naji A, Avrahami D, Golson ML &
Kaestner KH (2016) Single cell transcriptomics of the human endocrine
pancreas. Diabetes: db160405
}
\usage{
data(trainClassExample_wang, package = 'scClassify')
}
\description{
An obejct of scClassifyTrainModel for Wang et al.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{features}
\alias{features}
\alias{features,scClassifyTrainModel-method}
\title{Accessors of features for scClassifyTrainModel}
\usage{
features(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
features of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
features(trainClassExample_xin)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{model}
\alias{model}
\alias{model,scClassifyTrainModel-method}
\title{Accessors of model for scClassifyTrainModel}
\usage{
model(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
model of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
model(trainClassExample_xin)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{scClassifyTrainModelList}
\alias{scClassifyTrainModelList}
\title{The scClassifyTrainModelList class}
\usage{
scClassifyTrainModelList(...)
}
\arguments{
\item{...}{scClassifyTrainModel objects}
}
\value{
A scClassifyTrainModelList object
}
\description{
The scClassifyTrainModelList class
}
\examples{

data("trainClassExample_xin")
data("trainClassExample_wang")
trainClassExampleList <- scClassifyTrainModelList(trainClassExample_xin,
trainClassExample_wang
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\docType{class}
\name{scClassifyTrainModelList-class}
\alias{scClassifyTrainModelList-class}
\title{An S4 class to stored a list of training models from scClassify}
\description{
An S4 class to stored a list of training models from scClassify
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/learningCurve.R
\name{getN}
\alias{getN}
\title{Function to get the required N given by the accuracy and
the learning curve model}
\usage{
getN(res, acc = 0.9)
}
\arguments{
\item{res}{model results returned by \code{learning_curve} function}

\item{acc}{accuracy that are quired}
}
\value{
sample size that are required
}
\description{
Function to get the required N given by the accuracy and
the learning curve model
}
\examples{
set.seed(2019)
n <- seq(20, 10000, 100)
accMat <- do.call(cbind, lapply(1:length(n), function(i){
tmp_n <- rep(n[i], 50)
y <- -2/(tmp_n^0.8) + 0.95 + rnorm(length(tmp_n), 0, 0.02)
}))
res <- learningCurve(accMat = accMat, n)
N <- getN(res, acc = 0.9)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\docType{data}
\name{trainClassExample_xin}
\alias{trainClassExample_xin}
\title{Subset of pretrained model of Xin et al.}
\format{
An object of class \code{scClassifyTrainModel} of length 1.
}
\source{
Xin Y, Kim J, Okamoto H, Ni M, Wei Y, Adler C, Murphy AJ, Yancopoulos GD,
Lin C & Gromada J (2016) RNA sequencing of single human islet cells reveals
type 2 diabetes genes. Cell Metab. 24: 608–615
}
\usage{
data(trainClassExample_xin, package = 'scClassify')
}
\description{
An obejct of scClassifyTrainModel for Xin et al.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/scClassifyTrainClass.R
\name{cellTypeTrain}
\alias{cellTypeTrain}
\alias{cellTypeTrain,scClassifyTrainModel-method}
\title{Accessors of cellTypeTrain for scClassifyTrainModel}
\usage{
cellTypeTrain(x)
}
\arguments{
\item{x}{A `scClassifyTrainModel` object.}
}
\value{
cellTypeTrain of the scClassifyTrainModel slot
}
\description{
Methods to access various components of the `scClassifyTrainModel` object.
}
\examples{

data(trainClassExample_xin)
cellTypeTrain(trainClassExample_xin)

}
